/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.fraclet.extensions.Membrane;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.api.ADLParserSupportItf;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.CompileSupportItf;
import org.objectweb.fractal.juliac.api.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.MembraneLoaderItf;
import org.objectweb.fractal.juliac.api.RuntimeClassNotFoundException;
import org.objectweb.fractal.juliac.conf.GeneratorMap;
import org.objectweb.fractal.juliac.conf.JDKLevel;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.conf.JulietLoader;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.ucf.UClassFactory;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedClassFactoryItf;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeWriter;

/**
 * A Java source code generator for Fractal components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class Juliac {
    
    /**
     * Juliac command line entry point.
     */
    public static void main( String[] args )
    throws IOException, InstantiationException, RuntimeClassNotFoundException {
        
        if( args.length == 0 ) {
            usage();
            return;
        }
        
        /*
         * Register the command line options and flags.
         */
        CmdLineArgs<CmdLineFlags,CmdLineOptions> cla =
            new CmdLineArgs<CmdLineFlags,CmdLineOptions>();
        CmdLineOptions[] options = CmdLineOptions.values();
        for (CmdLineOptions option : options) {
            cla.registerOption(option);
            String[] domain = option.getDomain();
            if( domain != null ) {
            	cla.setOptionDomain(option,domain);
            }
		}
        cla.registerFlags(CmdLineFlags.values());
        
        Juliac jc = new Juliac();
        JuliacConfig jconf = new JuliacConfig(jc);
        Logger logger = jconf.getLogger();
        jc.setJuliacConfig(jconf);
        
        /*
         * Analyze the command line arguments.
         */
        try {
            cla.parse(args);
        }
        catch( IllegalArgumentException iae ) {
            String msg = iae.getMessage();
            logger.severe(msg);
            return;
        }

        /*
         * Retrieve the association between ASM bytecode generators and
         * source code generators for component interceptors.
         */
        String iclgstr = cla.getOptionValue(CmdLineOptions.ICLASSGS);
        if( iclgstr != null ) {
        	GeneratorMap<InterceptorClassGenerator> iclgs =
            		jconf.interceptorClassGenerators();
        	iclgs.set(iclgstr);
        }

        /*
         * Retrieve the association between ASM bytecode generators and
         * source code generators for component interceptors.
         */
        String iscgstr = cla.getOptionValue(CmdLineOptions.ICODEGS);
        if( iscgstr != null ) {
        	GeneratorMap<InterceptorSourceCodeGeneratorItf> iscgs =
            		jconf.interceptorSourceCodeGenerators();
        	iscgs.set(iscgstr);
        }

        /*
         * Retrieve other options and flags.
         * Assume default values when missing.
         */
        String s = cla.getOptionValue(CmdLineOptions.SRCS);
        String[] srcs = s==null ? new String[0] : s.split(File.pathSeparator);
        jconf.addSrcs(srcs);
        
        s = cla.getOptionValue(CmdLineOptions.MIXINS);
        String[] srclibs = s==null ? new String[0] : s.split(File.pathSeparator);
        jconf.addSrclibs(srclibs);
        
        String pkgRoot = cla.getOptionValue(CmdLineOptions.PKGROOT);
        if( pkgRoot != null ) {
            jconf.setPkgRoot(pkgRoot);
        }
        
        String genDirName = cla.getOptionValue(CmdLineOptions.GENSRC);
        if( genDirName != null ) {
            jconf.setGenDirName(genDirName);
        }
        
        String classDirName = cla.getOptionValue(CmdLineOptions.GENCLASS);
        if( classDirName != null ) {
            jconf.setClassDirName(classDirName);
        }
        
        String source = cla.getOptionValue(CmdLineOptions.SOURCE);
        if( source != null ) {
        	JDKLevel level = JDKLevel.valueOf(source);
            jconf.setSourceLevel(level);
        }
        
        String target = cla.getOptionValue(CmdLineOptions.TARGET);
        if( target != null ) {
        	JDKLevel level = JDKLevel.valueOf(target);
            jconf.setTargetLevel(level);
        }
        
        String classJarFileName = cla.getOptionValue(CmdLineOptions.GENCLASSJAR);
        if( classJarFileName != null ) {
            if( ! cla.isFlagSet(CmdLineFlags.COMPILEGENERATED) ) {
            	String msg = "--genclassjar cannot be used without --compiledGenerated";
            	logger.severe(msg);
            	return;
            }
            jconf.setClassJarFileName(classJarFileName);
        }
        
        boolean debug = cla.isFlagSet(CmdLineFlags.DEBUG);
        if(debug) {
        	JuliacConfig.setLevel(logger,Level.FINE);
        }
        
        boolean compilationWarnings =
        	cla.isFlagSet(CmdLineFlags.COMPILATIONWARNINGS);
        jconf.setCompilationWarnings(compilationWarnings);
        
        /*
         * Load modules last as the init procedure of some modules may depend
         * on configuration options and flags. E.g. this is the case for the
         * Spoon module which depends on --srcs and --mixins.
         */
        s = cla.getOptionValue(CmdLineOptions.MODULES);
        String[] modules = s==null ? new String[0] : s.split(":");
    	for (String module : modules) {
			jconf.loadModule(module);
		}
        
        /*
         * Retrieve the ADL files or membrane descriptors which must be
         * generated.
         */
        List<String> files = cla.getFiles();
        
        /*
         * Display debug information about the classpath if requested.
         */
    	logger.fine("Juliac classpath:");
        ClassLoader cl = Juliac.class.getClassLoader();
        if( cl instanceof URLClassLoader ) {
            URLClassLoader urlcl = (URLClassLoader) cl;
            URL[] urls = urlcl.getURLs();
            for (URL url : urls) {
            	logger.fine("  "+url);
            }
        }
        else {
        	String msg =
        		"Class loader for the Juliac class is not a URLClassLoader";
        	logger.fine(msg);
        }
        logger.fine("-----------------");
        
        /*
         * If requested, compile the code which is provided as input.
         */
        if( cla.isFlagSet(CmdLineFlags.COMPILEINPUT) ) {
            jc.compile();
        }
        
        /*
         * Generate the source code.
         */
        for (String file : files) {
        	logger.info(file+"...");
            
            /*
             * file may be either a membrane descriptor, a ADL descriptor, or a
             * list of component descriptors. Handle membrane descriptors first
             * in case the ADL and component descriptors use it.
             */
        	try {
                // file is a membrane descriptor
                jc.generate(file);
        	}
        	catch( IllegalArgumentException iae ) {
        		try {
            		// file is an ADL descriptor
                    jc.generate(file,file);
        		}
        		catch( IllegalArgumentException iae2 ) {
                	// file is a list of component descriptors
                	String[] descs = file.split(":");
            		for (String desc : descs) {
						jc.generateForDesc(desc);
					}
        		}
        	}
        }
        
        /*
         * If requested, compile the generated code.
         */
        if( cla.isFlagSet(CmdLineFlags.COMPILEGENERATED) ) {
        	jconf.setCompilationWarnings(false);
        	jc.compile();
        	jconf.setCompilationWarnings(compilationWarnings);
        }
        
        /*
         * If requested, package the compiled generated code.
         */
        if( classJarFileName != null ) {
        	jc.packGeneratedFiles();
        }
        
        /*
         * Close Juliac.
         */
        jc.close();
        logger.info("Done.");
    }
    
    /**
     * Display command line usage.
     */
    private static void usage() {
        
    	Juliac jc = new Juliac();
    	JuliacConfig jconf = new JuliacConfig(jc);
    	Logger logger = jconf.getLogger();
    	
        logger.info("java "+Juliac.class.getName()+" [options] types");
        logger.info("Generate the Java source code associated with the specified ADL, membrane or component descriptors");
        logger.info("");

        logger.info("Options:");
        logger.info("");
        logger.info("  --gensrc <dir>          : directory for generated source code (default: target/generated-sources/juliac)");
        logger.info("  --genclass <dir>        : directory for compiled code (default: target/classes)");
        logger.info("  --genclassjar <file>    : package compiled generate code in file");
        logger.info("  --iclassgs <generators> : interceptor class generators");
        logger.info("  --icodegs <generators>  : interceptor source code generators");
        logger.info("  --mixins <dirs>         : mixin layers");
        logger.info("  --modules <modules>     : modules");
        logger.info("  --pkgRoot <name>        : root package for generated code");
        logger.info("  --source <version>      : the Java source code compatibility level (default: 1.5)");
        logger.info("  --srcs <dirs>           : source files (default: src/main/java)");
        logger.info("  --target <version>      : the Java bytecode compatibility level (default: 1.5)");
        logger.info("");
        logger.info("  --compilationWarnings  : report compilation warnings");
        logger.info("  --compileInput         : compile input code");
        logger.info("  --compileGenerated     : compile generated code");
        logger.info("  --debug                : report debug information");
        logger.info("");

        logger.info("Details:");
        
        logger.info("  <generators> is a colon-separated list of pairs asmgen=srcgen");
        logger.info("    - asmgen is the fqname of the ASM bytecode generator");
        logger.info("    - srcgen is the fqname of the corresponding source code generator");
        logger.info("      e.g. org.objectweb.fractal.julia.asm.LifeCycleCodeGenerator=org.objectweb.fractal.juliac.proxy.LifeCycleSourceCodeGenerator");

        logger.info("  <dirs> is a colon-separated list of directories or jar files");
        
        logger.info("  <modules> is a colon-separated list of either");
        logger.info("    - either one of the predefined modules: JDT, JDK6, SPOON, FRACTAL_ADL, MIND_ADL, OO, COMP, MERGE_ALL, ULTRA_COMP");
        logger.info("    - or the fully-qualified name of a class that implements JuliacModuleItf");
    }
    

    // ----------------------------------------------------------------------
    // Singleton pattern, constructor and close method
    // ----------------------------------------------------------------------
    
    public Juliac() {
    	// Register the default {@link UnifiedClass} factory
        UnifiedClassFactoryItf ucf = new UClassFactory(this);
        register(ucf);
    }
    
    /**
     * @return  the {@link JuliacConfig} instance associated with this instance
     * @throws IllegalArgumentException
     *      if {@link #setJuliacConfig(JuliacConfig)} has not been called
     */
    public JuliacConfig getJuliacConfig() throws IllegalArgumentException {
        if( jconf == null ) {
            String msg =
                "JuliacConfig (jconf) has not been set for "+
                "this instance of Juliac: "+this;
            throw new IllegalArgumentException(msg);
        }
        return jconf;
    }
    public void setJuliacConfig( JuliacConfig jconf ) { this.jconf = jconf; }
    private JuliacConfig jconf;
    
    /**
     * Close the resources and discard the singleton instance of Juliac.
     */
    public void close() throws IOException {
    	if( jconf != null ) {
			jconf.unloadModules();
    	}
    }

    
    // ----------------------------------------------------------------------
    // Methods for generating component source code
    // ----------------------------------------------------------------------
    
    /**
     * Return the registered ADL parser that accepts the specified ADL
     * descriptor.
     * 
     * @param adl  the fully-qualified name of the ADL descriptor
     * @return     the ADL parser that accepts the specified ADL descriptor
     * @throws IllegalArgumentException
     * 				if none of the registered ADL parsers accept the
     * 				specified ADL descriptor
     * @since 2.3.1
     */
    public ADLParserSupportItf getADLDesc( String adl )
	throws IllegalArgumentException {
    	
    	JuliacConfig jconf = getJuliacConfig();
    	List<ADLParserSupportItf> adlParsers =
    		jconf.lookup(ADLParserSupportItf.class);
    	
    	for (ADLParserSupportItf adlParser : adlParsers) {
        	boolean b = adlParser.acceptADLDesc(adl);
        	if(b) {
	        	return adlParser;
        	}
		}

		String msg = "No such ADL parser for: "+adl;
		throw new IllegalArgumentException(msg);
    }
    
    /**
     * Generate the source code of a factory class for the specified ADL.
     * 
     * @param adl         the fully-qualified name of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     * @throws IllegalArgumentException
     * 				if none of the registered ADL parsers accept the
     * 				specified ADL descriptor
     * @since 2.3.1
     */
    public void generate( String adl, String targetname )
    throws IOException, IllegalArgumentException {
    	
		ADLParserSupportItf supportItf = getADLDesc(adl);
        String rootedName = getRootedClassName(targetname);

        JuliacConfig jconf = getJuliacConfig();
        Logger logger = jconf.getLogger();
        logger.info("[ADLParser: "+supportItf.getClass().getName()+"]");
        
        supportItf.generate(adl,rootedName);
    }
    
    /**
     * Generate the source code for the specified membrane descriptor.
     * 
     * @param controllerDesc  the membrane descriptor
     * @return  the source code generator
     * @throws IllegalArgumentException  if no source code generator is suitable
     */
    public FCSourceCodeGeneratorItf<?> generate( Object controllerDesc )
    throws IOException, InstantiationException, IllegalArgumentException {
        
        String ctrlDesc = (String) controllerDesc;
        ComponentType ct = new BasicComponentType(new BasicInterfaceType[0]);

        FCSourceCodeGeneratorItf<?> fcscg = null;
        try {
        	fcscg = getFCSourceCodeGenerator(ctrlDesc);
        }
        catch( IllegalArgumentException iae ) {
        	
        	/*
        	 * No source code generator for this membrane descriptor.
        	 * Check whether the descriptor is a @Membrane annotated class.
        	 * If so, register the class with all existing JulietLoaders.
        	 * 
        	 */
        	JuliacConfig jconf = getJuliacConfig();
        	Class<?> cl = null;
        	try {
        		cl = jconf.load(ctrlDesc);
        	}
        	catch( RuntimeClassNotFoundException rcnfe ) {
        		String msg = iae.getMessage();
        		throw new IllegalArgumentException(msg);
        	}
        	Membrane membrane = cl.getAnnotation(Membrane.class);
        	ctrlDesc = membrane.desc();
	    	List<JulietLoader> mloaders = jconf.lookup(JulietLoader.class);
	    	for( MembraneLoaderItf mloader : mloaders ) {
    			mloader.put(ctrlDesc,cl);
	    	}
	    	
	    	/*
	    	 * Now we should be able to retrieve a source code generator that
	    	 * accepts the specified membrane descriptor. All generators
	    	 * associated with a JulietLoader match. Conversely, it may be the
	    	 * case that no JulietLoader has been registered, and in this case,
	    	 * no generator matches.
	    	 */
	    	fcscg = getFCSourceCodeGenerator(ctrlDesc);
        }

        JuliacConfig jconf = getJuliacConfig();
        Logger logger = jconf.getLogger();
        logger.info("[Generator: "+fcscg.getClass().getName()+"]");
        
		fcscg.generateMembraneImpl(ct,ctrlDesc,null);
        return fcscg;
    }
    
    /**
     * Generate the source code for the specified proxy interface and membrane
     * descriptor.
     * 
     * @param signature       the interface signature
     * @param controllerDesc  the membrane descriptor
     * @return  the source code generator
     * @throws IllegalArgumentException  if no source code generator is suitable
     */
    public FCSourceCodeGeneratorItf<?> generateProxiesImpl(
		String signature, Object controllerDesc )
    throws IOException, IllegalArgumentException {
        
        String ctrlDesc = (String) controllerDesc;
        FCSourceCodeGeneratorItf<?> fcscg = getFCSourceCodeGenerator(ctrlDesc);
        
        JuliacConfig jconf = getJuliacConfig();
        Logger logger = jconf.getLogger();
        logger.info("[Generator: "+fcscg.getClass().getName()+"]");

        fcscg.generateProxiesImpl(signature,controllerDesc);
        return fcscg;
    }
    
    /**
     * Generate the source code for the specified component.
     * 
     * @param type            the component type
     * @param controllerDesc  the controller descriptor
     * @param contentDesc     the content descriptor
     * @return  the initializer class generator
     * @throws IllegalArgumentException  if no source code generator is suitable
     * @since 2.2.4
     */
    public InitializerClassGenerator<?> generate(
		Type type, Object controllerDesc, Object contentDesc )
    throws IOException, IllegalArgumentException {

    	FCSourceCodeGeneratorItf<?> fcscg =
        	getFCSourceCodeGenerator(controllerDesc);
        
    	JuliacConfig jconf = getJuliacConfig();
        Logger logger = jconf.getLogger();
        logger.info("[Generator: "+fcscg.getClass().getName()+"]");
        
        InitializerClassGenerator<?> icg =
        	fcscg.generate(type,controllerDesc,contentDesc);
        return icg;
    }
    
    /**
     * Generate the source code for the specified component descriptor.
     * 
     * The descriptor is expected to be of the form
     * <code>[name/signature/bool/bool/bool;...,controllerDesc,contentDesc]</code>.
     * 
     * @param desc  the component descriptor
     * @throws IllegalArgumentException
     * 				if the descriptor does not match the expected syntax or
     * 				if no source code generator is suitable
     * @since 2.2.4
     */
    public void generateForDesc( String desc )
    throws IOException, IllegalArgumentException, InstantiationException {
    	
    	/*
    	 * This method is named generateForDesc instead of generate to avoid
    	 * mismatches with the generate(Object) method which was existing prior
    	 * to version 2.2.4. Else pieces of code which were using
    	 * generate(Object) while transmitting a String would end up executing
    	 * instead generate(String).
    	 */
    	
		final String msg =
			"Expected component descriptor "+
			"[name/signature/bool/bool/bool;...,controllerDesc,contentDesc]. "+
			"Got: "+desc;

		if( desc.length() < "[x/x/true/true/true,x,x]".length() ) {
			throw new IllegalArgumentException(msg);
		}

		if( desc.charAt(0)!='[' && desc.charAt(desc.length()-1)!=']' ) {
			throw new IllegalArgumentException(msg);
    	}
		desc = desc.substring(1,desc.length()-1);
    	
		/*
		 * We need 3 arguments delimited by a comma or 2 if the trailing
		 * character is a comma.
		 */
		String[] elts = desc.split(",");
		if( !( elts.length==3 || (elts.length==2 && desc.endsWith(",")) ) ) {
			throw new IllegalArgumentException(msg);
		}
		String t = elts[0];
		String ctrlDesc = elts[1];
		String contentDesc = elts.length==2 ? null: elts[2];
		
		String[] types = t.split(";");
		InterfaceType[] its = new InterfaceType[types.length];
		for( int i=0 ; i<types.length ; i++ ) {
			String type = types[i];
			elts = type.split("/");
			if( elts.length != 5 ) {
				throw new IllegalArgumentException(msg);
			}
			String name = elts[0];
			String signature = elts[1];
			boolean isClient = Boolean.parseBoolean(elts[2]);
			boolean isOptional = Boolean.parseBoolean(elts[3]);
			boolean isCollection = Boolean.parseBoolean(elts[4]);
			its[i] = 
				new BasicInterfaceType(
					name, signature, isClient, isOptional, isCollection );
		}
		ComponentType ct = new BasicComponentType(its);
		
		generate(ct,ctrlDesc,contentDesc);
    }

    /**
     * Dump the code generated by the specified source code generator in its
     * corresponding file. Do not perform the generation if the generated type
     * exists in the classpath.
     */
    public void generateSourceCode( ClassGeneratorItf scg ) throws IOException {
        
        /*
         * Avoid generating code for existing types.
         */
        String targetClassName = scg.getTargetClassName();
        ClassLoader classLoader = getJuliacConfig().getClassLoader();
        try {
            classLoader.loadClass(targetClassName);
            return;
        }
        catch( ClassNotFoundException cnfe ) {}
        
        generateSourceCodeOverride(scg);
    }
    
    /**
     * Dump the code generated by the specified source code generator in its
     * corresponding file. Perform the generation even if the generated type
     * already exists in the classpath.
     */
    public void generateSourceCodeOverride( ClassGeneratorItf scg )
    throws IOException {
        
        JuliacConfig jconf = getJuliacConfig();
        File genDir = jconf.getGenDir();
        String targetClassName = scg.getTargetClassName();
        CompilationRounds rounds = getJuliacConfig().getCompilationRounds();

        /*
         * Avoid duplicate generation.
         */
        SourceFile sf = SourceFile.create(genDir,targetClassName);
        if( rounds.contains(sf) ) {
            return;
        }
        
        String filename =
            targetClassName.replace('.',File.separatorChar)+".java";
        File f = new File(genDir,filename);
        File fdir = f.getParentFile();
        fdir.mkdirs();

        /*
         * Do not generate directly in a FileWriter to avoid generating
         * incomplete files which may cause errors when trying to compile.
         * 
         * The issue appeared with the <gracefulExit> MOJO option. The purpose
         * is to exit the MOJO without reporting an exception (the functionality
         * was requested by Nicolas from Petals Link.) If the generation throws
         * an exception, an incomplete file may be generated, the MOJO exits,
         * and the regular Maven compiler MOJO would try to compile the
         * incomplete file. Of course, an error would be reported.
         * 
         * To prevent this from happening, we first generate the code in memory
         * (StringWriter) and then dump the content in the real file once the
         * generation has been completed.
         */
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        FileSourceCodeVisitor cw = new FileSourceCodeWriter(pw);
        scg.generate(cw);
        pw.close();
        sw.close();
    
        FileWriter fw = new FileWriter(f);
        StringBuffer sb = sw.getBuffer();
        fw.write(sb.toString());
        fw.close();

        CompilationRound round = rounds.getCurrentCompilationRound();
        round.addGenerated(SourceFile.create(genDir,targetClassName));
    }

    /**
     * Return <code>true</code> if the specified type has already been generated
     * or exists in the classpath.
     */
    public boolean hasBeenGenerated( String name ) throws IOException {
        
        JuliacConfig jconf = getJuliacConfig();
        File genDir = jconf.getGenDir();
        CompilationRounds rounds = jconf.getCompilationRounds();

        // First check whether the type has been generated by Juliac
        SourceFile igf = SourceFile.create(genDir,name);
        if( rounds.contains(igf) ) {
            return true;
        }
        
        // Check whether the type exists in the classpath
        ClassLoader classLoader = jconf.getClassLoader();
        try {
            classLoader.loadClass(name);
            return true;
        }
        catch( ClassNotFoundException cnfe ) {}
        
        return false;
    }
    
    /**
     * If a package root is defined by the {@link JuliacConfig} instance
     * associated with the current instance, return a name where the package
     * root is appended in front of the specified content class name. Else
     * return the content class name unchanged.
     */
    public String getRootedClassName( String contentClassName ) {
        
        JuliacConfig jconf = getJuliacConfig();
        String pkgRoot = jconf.getPkgRoot();
        
        if( pkgRoot!=null && pkgRoot.length()!=0 ) {
            StringBuffer rootedClassName = new StringBuffer();
            rootedClassName.append(pkgRoot);
            rootedClassName.append('.');
            rootedClassName.append(contentClassName);
            return rootedClassName.toString();
        }
        else {
            return contentClassName;
        }
    }

    /**
     * Return the source code generator which can generate the source code for
     * the specified parameters. If multiple generators are suitable, return the
     * first one.
     * 
     * @param controllerDesc  the component controller descriptor
     * @return  the source code generator which can generate the code for the
     *          specified controller descriptor
     * @throws IllegalArgumentException  if no source code generator is suitable
     */
	public FCSourceCodeGeneratorItf<?> getFCSourceCodeGenerator(
        Object controllerDesc )
    throws IllegalArgumentException {
        
        JuliacConfig jconf = getJuliacConfig();
        
        String ctrlDesc = (String) controllerDesc;
        @SuppressWarnings("rawtypes")
		List<FCSourceCodeGeneratorItf> fcscgs =
        	jconf.lookup(FCSourceCodeGeneratorItf.class);

        /*
         * Apply a chain of responsibility pattern on all registered source code
         * generators.
         */
        for (FCSourceCodeGeneratorItf<?> fcscg : fcscgs) {
            boolean accept = fcscg.acceptCtrlDesc(ctrlDesc);
            if( accept ) {
                return fcscg;
            }
        }
        
        /*
         * No suitable source code generator.
         */
        String msg =
            "No source code generator for controller descriptor: "+
            controllerDesc;
        throw new IllegalArgumentException(msg);
    }

    
    // ----------------------------------------------------------------------
    // Methods for compiling and packaging code
    // ----------------------------------------------------------------------

    /**
     * Compile the source code contained in the current compilation round. The
     * source code which is compiled may be either input source code or
     * generated source code.
     * 
     * @return  the compilation round which contains the compiled files
     * @throws IOException  in case of compilation error
     * @since 2.0
     */
    public CompilationRound compile() throws IOException {
        
        JuliacConfig jconf = getJuliacConfig();
    	Logger logger = jconf.getLogger();
    	logger.info("Compiling...");
    	
        /*
         * First check if there are some files to compile.
         */
        CompilationRounds rounds = jconf.getCompilationRounds();
        CompilationRound round = rounds.getCurrentCompilationRound();
        List<SourceFile> inputFiles = round.getInputFiles();
        List<SourceFile> generatedFiles = round.getGeneratedFiles();
        if( inputFiles.size()==0 && generatedFiles.size()==0 ) {
        	logger.info("No file to compile");
            return round;
        }
        
        /*
         * Set the compilation log level.
         */
        Logger compilationLogger = jconf.getCompilationLogger();
        boolean compilationWarnings = jconf.getCompilationWarnings();
        if(compilationWarnings) {
        	JuliacConfig.setLevel(compilationLogger,Level.WARNING);
        }
        
        /*
         * Compile the files.
         */
        CompileSupportItf compiler =
        	jconf.lookupUnique(CompileSupportItf.class);
        round.compile(compiler);

        /*
         * Reset the compilation log level to log errors only. Needed in case
         * the compilation is performed with the #compileQuiet() method.
         */
    	JuliacConfig.setLevel(compilationLogger,Level.SEVERE);
        
        /*
         * Report on compiled files.
         */
        for (SourceFile isf : inputFiles) {
        	logger.fine("   "+isf.getQname());
        }
        for (SourceFile igf : generatedFiles) {
        	logger.fine("   "+igf.getQname());
        }
        int total = inputFiles.size()+generatedFiles.size();
        String msg =
            total+" file(s) compiled to "+
            round.getClassDir().getAbsolutePath();
        logger.info(msg);
        
        return round;
    }
    
    /**
     * Compile the source code contained in the current compilation round. The
     * source code which is compiled may be either input source code or
     * generated source code. Only errors are reported.
     * 
     * @return  the compilation round which contains the compiled files
     * @throws IOException  in case of compilation error
     * @since 2.2.4
     */
    public CompilationRound compileQuiet() throws IOException {
    	
    	/*
    	 * Copy/paste of the #compile() method without log invocations. Getting
    	 * the current log level and re-installing it hadn't been successful
    	 * (the current log level returned by logger.getLogLevel() seemed to be
    	 * null?). And by the way, this wouldn't work in case of concurrent
    	 * compilation even if such a case is unlikely to occur given the
    	 * current usage patterns of Juliac (but who knows in the future ;-)
    	 */
    	
        CompilationRounds rounds = jconf.getCompilationRounds();
        CompilationRound round = rounds.getCurrentCompilationRound();

        CompileSupportItf compiler =
        	jconf.lookupUnique(CompileSupportItf.class);
        round.compile(compiler);

        return round;
    }
    
    /**
     * Package the compiled generated files.
     * 
     * @since 2.2.4
     */
    public void packGeneratedFiles() throws IOException {
    	
    	JuliacConfig jconf = getJuliacConfig();
    	Logger logger = jconf.getLogger();
    	CompilationRounds rounds = jconf.getCompilationRounds();
    	List<CompilationRound> rds = rounds.getCompilationRounds();
    	        
        // Check that some files have been generated
    	boolean generated = false;
    	for (CompilationRound round : rds) {
            List<String> outputFileNames = round.getOutputFileNames();
    		if( outputFileNames.size() != 0 ) {
    			generated = true;
    			break;
    		}
    	}
        if( ! generated ) {
        	String msg =
        		"No compiled generated file. Cannot create the jar file.";
        	logger.warning(msg);
        	return;
        }
        
        // Create the jar file
    	File jarFile = jconf.getClassJarFile();
    	OutputStream os = new FileOutputStream(jarFile); 
    	JarOutputStream jos = new JarOutputStream(os);
    	int total = 0;

    	// Iterate on each compilation round
    	for (CompilationRound round : rds) {
			
            File classDir = round.getClassDir();
            List<String> outputFileNames = round.getOutputFileNames();

            // Iterate on each generated compiled file contained in the round
            for (String outputFileName : outputFileNames) {
    			
        		// Retrieve the output file
        		File f = new File(classDir,outputFileName);
    			FileInputStream fis = new FileInputStream(f);
    			long size = fis.available();

    			// Create a new entry in the jar file
    			JarEntry je = new JarEntry(outputFileName);
    			je.setSize(size);
        		jos.putNextEntry(je);
        		
        		// Dump the content of the output file in the jar file
        		byte[] b = new byte[(int)size];
        		fis.read(b);
        		jos.write(b);
        		
        		// Close the entry in the jar file
        		jos.closeEntry();
        		total++;
    		}
		}    	
    	    	
    	// Close the jar file
    	jos.close();
    	os.close();
    	
    	// Report
    	String msg = total+" file(s) packaged to "+ jarFile.getAbsolutePath();
        logger.info(msg);
    }

    
    // ----------------------------------------------------------------------
    // Management of UnifiedClass factories.
    // ----------------------------------------------------------------------

    public void register( UnifiedClassFactoryItf ucf ) {
        ucfs.add(ucf);
    }
   
    public UnifiedClass create( String name )
	throws RuntimeClassNotFoundException {
        
        /*
         * Chain of responsibility pattern on the factories which are
         * registered.
         */
        for (UnifiedClassFactoryItf ucf : ucfs) {
            try {
                UnifiedClass uc = ucf.create(name);
                return uc;
            }
            catch( RuntimeClassNotFoundException cnfe ) {}
        }
        
        throw new RuntimeClassNotFoundException(name);
    }

    public String getInterfaceTypeSignature( InterfaceType it ) {
        
        String signature = it.getFcItfSignature();
        
        UnifiedClass proxycl = create(it.getFcItfSignature());
        String[] tpnames = proxycl.getTypeParameterNames();
        if( tpnames.length == 0 ) {
            return signature;
        }
        
        StringBuffer sb = new StringBuffer(signature);
        String s = Utils.getTypeParameterNamesSignature(tpnames);
        sb.append(s);
        return sb.toString();
    }

    /** Registered {@link UnifiedClass} factories. */
    private List<UnifiedClassFactoryItf> ucfs =
        new ArrayList<UnifiedClassFactoryItf>();
}
