/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac;

import java.io.CharArrayWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

/**
 * This class represents a source code file, either provided as input or
 * generated, which is manipulatd by Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.3
 */
public abstract class SourceFile {
    
    protected File root;
    protected String qname;
    
    /**
     * Factory method for creating {@link SourceFile}s.
     * 
     * @param root
     *      the root from which the source file is stored (directory or .jar
     *      file)
     * @param qname
     *      the fully-qualified name of the type contained in the source file
     * @return
     *      the corresponding {@link SourceFile}
     * @throws IllegalArgumentException
     *      if <code>root</code> is neither a directory, nor a .jar file
     */
    public static SourceFile create( File root, String qname )
    throws IllegalArgumentException {
        
        if( root.isDirectory() ) {
            SourceFile sf = new SourceFileDir(root,qname);
            return sf;
        }
        
        String name = root.getName();
        if( name.endsWith(".jar") ) {
            SourceFile sf = new SourceFileJar(root,qname);
            return sf;
        }
        
        String msg = root+" should be either a directory or a .jar file";
        throw new IllegalArgumentException(msg);
    }
    
    /**
     * Factory method for creating virtual {@link SourceFile}s whose content is
     * stored in a String.
     * 
     * @param qname
     *      the fully-qualified name of the type contained in the source file
     * @param content  the content
     * @return         the corresponding {@link SourceFile}
     */
    public static SourceFile create( String qname, String content ) {
        SourceFile sf = new SourceFileString(qname,content);
        return sf;
    }
    
    protected SourceFile( File file, String qname ) {
        this.root = file;
        this.qname = qname;
    }
    
    
    // -----------------------------------------------------------------------
    // Public methods
    // -----------------------------------------------------------------------
        
    /**
     * Return the content of the source file.
     */
    public abstract char[] getContent() throws IOException;
    
    /**
     * Return the file name of the source file.
     */
    public String getName() {
        String[] names = qname.split("[.]");
        String name = names[names.length-1]+".java";
        return name;
    }
    
    /**
     * Add all Java files contained in the specified location to the specified
     * list of {@link SourceFile}s. A location denotes either an absolute path
     * or a path relative to <code>baseDir</code>, for a directory or a jar
     * file.
     * 
     * @param baseDir  the base directory
     * @param src      the location
     * @param sfs      the list of source files
     * @since 2.2.1
     */
    public static void addAllJavaFiles(
        File baseDir, String src, List<SourceFile> sfs )
    throws IOException {
        
        File input = new File(baseDir,src);
            
        if( ! input.exists() ) {
            // Handle the case where inputName is an absolute path
            input = new File(src);
            if( ! input.exists() ) {
                String msg = "No such file or directory <"+baseDir+","+src+">";
                throw new IOException(msg);
            }
        }
        
        if( input.isDirectory() ) {
            SourceFile.addAllJavaFilesInDir(input,input,sfs);
        }
        else {
            String name = input.getName();
            String ext = name.substring(name.length()-4);
            if( ext.equals(".jar") ) {
                SourceFile.addAllJavaFilesInJar(input,sfs);
            }
            else {
                String msg = src+ " should be a directory or a .jar file";
                throw new IOException(msg);
            }
        }
    }
    
    /**
     * Add all Java source files contained in the specified <code>dir</code>
     * directory to the specified list of {@link SourceFile}s. Recursively scan
     * the directory.
     * 
     * @param root  the root directory where the scan begins
     * @param dir   the current directory to scan
     * @param sfs   the list of source files
     */
    private static void addAllJavaFilesInDir(
        File root, File dir, List<SourceFile> sfs )
    throws IOException {
        
        File[] fs = dir.listFiles();
        if( fs == null ) {
            String msg =
                dir.getAbsolutePath()+" was supposed to be a directory";
            throw new IOException(msg);
        }
        
        /*
         * Handle .java files contained in the current directory.
         */
        int headlength = root.getAbsolutePath().length();
        for (File f : fs) {
            if( ! f.isDirectory() && f.getName().endsWith(".java") ) {
                String fname = f.getAbsolutePath();
                fname = fname.substring(headlength+1);  // +1 for /
                fname = fname.substring(0,fname.length()-5);  // remove ".java"
                fname = fname.replace(File.separatorChar,'.');
                SourceFile sf = SourceFile.create(root,fname);
                sfs.add(sf);
            }
        }
        
        /*
         * Recursively scan subdirectories.
         */
        for (File f : fs) {
            if( f.isDirectory() ) {
                addAllJavaFilesInDir(root,f,sfs);
            }
        }
    }

    /**
     * Add all Java source files contained in the specified .jar file to the
     * specified list of {@link SourceFile}s.
     * 
     * @param jar  the jar file
     * @param sfs  the list of source files
     */
    private static void addAllJavaFilesInJar( File jar, List<SourceFile> sfs )
    throws IOException {
        
        FileInputStream fis = new FileInputStream(jar);
        JarInputStream jis = new JarInputStream(fis);
        
        JarEntry je = jis.getNextJarEntry();
        while( je != null ) {
            
            String name = je.getName();
            if( je.isDirectory() || ! name.endsWith(".java") ) {
                // Skip non .java files
                je = jis.getNextJarEntry();
                continue;
            }
            
            // Entry qname
            String fname = name.substring(0,name.length()-5);  // remove ".java"
            fname = fname.replace('/','.');

            /*
             * Entry content.
             * 
             * je.getSize() proved to be unsuccesful on some jar files created
             * with the command line jar cf. Returned size was -1. Switch to
             * method read(byte[],int,int) which returns -1 when the end of the
             * entry is reached. Use exactly this method (and not one of the
             * other read methods) as this is the only one which is overriden
             * in JarInputStream.
             */
            CharArrayWriter caw = new CharArrayWriter();
            byte[] buffer = new byte[1024];
            int length;
            while( (length=jis.read(buffer,0,buffer.length)) != -1 ) {
                for (int i = 0; i < length; i++) {
                    char c = (char) buffer[i];
                    caw.append(c);
                }
            }
            char[] content = caw.toCharArray();
            
            SourceFileJar sfj = new SourceFileJar(jar,fname);
            sfj.setContent(content);
            sfs.add(sfj);
            
            je = jis.getNextJarEntry();
        }
        
        jis.close();
        fis.close();
    }
    
    
    // -----------------------------------------------------------------------
    // Getters
    // -----------------------------------------------------------------------
    
    /**
     * Return the fully-qualified name of the type contained in the source file.
     */
    public String getQname() {
        return qname;
    }

    /**
     * Return the root from which the source file is stored (directory or .jar
     * file).
     */
    public File getRoot() {
        return root;
    }
    
    
    // -----------------------------------------------------------------------
    // j.l.Object methods
    // -----------------------------------------------------------------------
    
    @Override
    public boolean equals( Object other ) {
        if( !(other instanceof SourceFile) ) {
            return false;
        }
        return qname.equals(((SourceFile)other).qname);
    }
    
    @Override
    public int hashCode() {
        return qname.hashCode();
    }
    
    @Override
    public String toString() {
        String s =
            root.getAbsolutePath() + File.separatorChar +
            qname.replace('.',File.separatorChar) + ".java";
        return s;
    }

    
    // -----------------------------------------------------------------------
    // Implementation of a source file stored in a directory
    // -----------------------------------------------------------------------
    
    /**
     * {@link SourceFile} stored in a directory.
     * 
     * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
     */
    private static class SourceFileDir extends SourceFile {
        
        protected SourceFileDir( File root, String qname ) {
            super(root,qname);
        }
        
        public char[] getContent() throws IOException {
            
            String fname = qname.replace('.',File.separatorChar)+".java";
            File f = new File(root,fname);
            FileReader fr = new FileReader(f);

            long length = f.length();
            int ilength = (int) length;
            char[] content = new char[ilength];
            fr.read(content);
            fr.close();
            
            return content;
        }
    }

    
    // -----------------------------------------------------------------------
    // Implementation of a source file stored in a .jar file
    // -----------------------------------------------------------------------
    
    /**
     * {@link SourceFile} stored in a jar file.
     * 
     * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
     */
    private static class SourceFileJar extends SourceFile {
        
        protected SourceFileJar( File root, String qname ) {
            super(root,qname);
        }
        
        protected void setContent( char[] content ) {
            this.content = content;
        }
        
        public char[] getContent() {
            return content;
        }
        
        private char[] content;
    }

    // -----------------------------------------------------------------------
    // Implementation of a source file stored in a String
    // -----------------------------------------------------------------------
    
    /**
     * {@link SourceFile} stored in a String.
     * 
     * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
     * @since 2.2.1
     */
    private static class SourceFileString extends SourceFile {
        
        protected SourceFileString( String qname, String content ) {
            super(new File("."),qname);
            this.content = content.toCharArray();
        }
        
        public char[] getContent() {
            return content;
        }
        
        private char[] content;
    }
}
