/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Valerio Schiavoni, Christophe Munilla
 */

package org.objectweb.fractal.juliac;

import java.io.File;
import java.util.Set;
import java.util.TreeSet;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.conf.Constants;

/**
 * This class provides general purpose helper methods for Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Valerio Schiavoni <valerio.schiavoni@gmail.com>
 * @author Christophe Munilla <Christophe.Munilla@inria.fr>
 */
public class Utils {

    /**
     * Return a string containing the Java code to create a content descriptor
     * similar to the specified parameter.
     */
    public static StringBuffer javaifyContentDesc( Object contentDesc ) {

        StringBuffer sb = new StringBuffer();
    
        if( contentDesc == null ) {
            sb.append("null");
        }
        else if( contentDesc instanceof Object[] ) {
            sb.append("new Object[]{");
            Object[] array = (Object[]) contentDesc;
            for (Object o : array) {
                sb.append(javaifyContentDesc(o));
                sb.append(',');
            }
            sb.append("}");
        }
        else {
            // Assume the only other possible case is String
            sb.append('"');
            sb.append(contentDesc);
            sb.append('"');
        }
        
        return sb;
    }

    /**
     * Return a string containing the Java code to create a {@link
     * ComponentType} similar to the specified parameter.
     */
    public static StringBuffer javaify( ComponentType ct ) {
        InterfaceType[] its = ct.getFcInterfaceTypes();
        return javaify(its);
    }

    /**
     * Return a string containing the Java code to create an array of {@link
     * InterfaceType} similar to the specified parameter.
     */
    public static StringBuffer javaify( InterfaceType[] its ) {
    
        StringBuffer sb = new StringBuffer();
    
        sb.append("new "+BasicComponentType.class.getName()+"( ");
        sb.append("new "+InterfaceType.class.getName()+"[]{");
        
        for (int i = 0; i < its.length; i++) {
            sb.append(javaify(its[i]));
            sb.append(",");
        }
        
        sb.append("} )");
    
        return sb;
    }

    /**
     * Return a string containing the Java code to create an {@link
     * InterfaceType} similar to the specified parameter.
     */
    public static StringBuffer javaify( InterfaceType it ) {
        
        StringBuffer sb = new StringBuffer();
    
        String itfname = it.getFcItfName();
        sb.append("new "+BasicInterfaceType.class.getName()+"(");
        sb.append("\""+itfname+"\",");
        sb.append("\""+it.getFcItfSignature()+"\",");
        sb.append((it.isFcClientItf()?"true":"false")+",");
        sb.append((it.isFcOptionalItf()?"true":"false")+",");
        sb.append((it.isFcCollectionItf()?"true":"false"));
        sb.append(")");
    
        return sb;
    }
    
    /**
     * Return a string containing the Java code to create a {@link Tree}
     * similar to the specified parameter.
     */
    public static StringBuffer javaify( Tree tree ) {

        StringBuffer sb = new StringBuffer();
        
        sb.append("new "+Tree.class.getName()+'(');
        if( tree.getSize() == 0 ) {
            sb.append('"'+tree.toString()+'"');
        }
        else {
            sb.append("new "+Tree.class.getName()+"[]{");
            Tree[] subtrees = tree.getSubTrees();
            boolean first = true;
            for (Tree subtree : subtrees) {
                if(first) {first=false;} else {sb.append(',');}
                StringBuffer s = javaify(subtree);
                sb.append(s);
            }
            sb.append('}');
        }
        sb.append(')');
        
        return sb;
    }

    /**
     * Return a copy of the given interface type where the role of the interface
     * has been inverted: if the role of the specified interface type is client
     * (resp. server), the role of the returned interface type is server (resp.
     * client).
     */
    public static InterfaceType newSymetricInterfaceType( InterfaceType it ) {
    	String name = it.getFcItfName();
        return newSymetricInterfaceType(it,name);
    }

    /**
     * Return a copy of the given interface type where the role of the interface
     * has been inverted: if the role of the specified interface type is client
     * (resp. server), the role of the returned interface type is server (resp.
     * client).
     * 
     * @param it    the interface type
     * @param name  the name of the new interface type
     * @since 2.5
     */
    public static InterfaceType newSymetricInterfaceType(
		InterfaceType it, String name ) {
        return
            new BasicInterfaceType(
                name,
                it.getFcItfSignature(),
                ! it.isFcClientItf(),   // client <-> server
                it.isFcOptionalItf(),
                it.isFcCollectionItf()
            );        
    }

    /**
     * Given the interface types of a control membrane, return the corresponding
     * type for the application (level 0) interfaces. This consists in removing
     * the two slash characters which are in front of the interface names.
     * 
     * @since 2.5
     */
    public static InterfaceType[] downToLevel0InterfaceType( InterfaceType[] srcits ) {
        InterfaceType[] its = new InterfaceType[ srcits.length ];
        for (int i = 0; i < srcits.length; i++) {
            its[i] = Utils.downToLevel0InterfaceType(srcits[i]);
        }
        return its;
    }
    
    /**
     * Given the interface type of a control membrane, return the corresponding
     * type for the application (level 0) interface. This consists in removing
     * the two slash characters which are in front of the interface name.
     */
    public static InterfaceType downToLevel0InterfaceType( InterfaceType srcit ) {
        InterfaceType it =
            new BasicInterfaceType(
                srcit.getFcItfName().substring(2),     // remove //
                srcit.getFcItfSignature(),
                srcit.isFcClientItf(),
                srcit.isFcOptionalItf(),
                srcit.isFcCollectionItf()
            );
        return it;
    }
    
    /**
     * Return <code>true</code> if one the specified {@link InterfaceType}s
     * defines a <code>content-controller</code> or a
     * <code>basic-content-controller</code> interface. This denotes a
     * composite component. <code>basic-content-controller</code> is defined by
     * Dream for the <code>dreamActivitiesComposite</code> controller
     * descriptor.
     * 
     * @since 2.1.6
     */
    public static boolean isComposite( InterfaceType[] its ) {
        for (InterfaceType it : its) {
            String itname = it.getFcItfName();
            if( itname.equals("content-controller") ||
                itname.equals("basic-content-controller") ) {
                return true;
            }
        }
        return false;
    }

    /**
     * Return a stringified hash code for the specified {@link ComponentType}.
     */
    public static String hexhash( ComponentType ct ) {
    	
    	/*
    	 * Sort interface types in the alphabetical order of their name. The
    	 * idea is that, since component types [I,J] and [J,I] are the same,
    	 * the same hash should be returned, in order to avoid duplicate
    	 * unecessary code generation.
    	 * 
    	 * This fix has been introduced in Juliac 2.4 on April 1, 2011,
    	 * following a discussion with Philippe that requests this feature in
    	 * FraSCAti when modules are merged to produce a component, but that the
    	 * merge order leads to different component types, although only one
    	 * component type exists (but with interfaces sorted in different
    	 * orders.) 
    	 */
    	
    	InterfaceType[] its = ct.getFcInterfaceTypes();
    	Set<String> names = new TreeSet<String>();
    	for (InterfaceType it : its) {
			String name = it.getFcItfName();
			names.add(name);
		}
    	
    	StringBuffer sct = new StringBuffer("[");
        boolean first = true;
        for (String name : names) {
			InterfaceType it = null;
			try {
				it = ct.getFcInterfaceType(name);
			}
			catch (NoSuchInterfaceException nsie) {
				// Shouldn't occur since name if one the interface names from ct
				throw new JuliacRuntimeException(nsie);
			}
            if(first) { first = false; }
            else { sct.append(','); }
            sct.append(it.toString());
        }
        sct.append(']');
    	
        /*
         * toString() mandatory to compute the hash on the String, not the
         * StringBuffer. 2 Strings with the same content have the same hash
         * code, whereas 2 StringBuffers with the same content are considered as
         * different objects, and do not share the same hash code.
         */
        int hashct = sct.toString().hashCode();
        String hexhashct = Integer.toHexString(hashct);
        return hexhashct;
    }

    /**
     * Return a usable name for a type which must be generated by Juliac.
     * Two categories of types are moved to {@link
     * Constants#JULIAC_RUNTIME_GENERATED}:
     * 
     * <ul>
     * <li>types with an empty package name,</li>
     * <li>types with a Java reserved package name (starting with java).</li>
     * </ul>
     * 
     * Other types are unchanged.
     * 
     * @param fqn  the fully-qualified name of the type which is generated
     * @return     a usable fully-qualified name for this type
     */
    public static String getJuliacGeneratedStrongTypeName( String fqn ) {
        
        int lastdot = fqn.lastIndexOf('.');
        
        // No package name
        if( lastdot == -1 ) {
            String s = Constants.JULIAC_RUNTIME_GENERATED + "." + fqn;
            return s;
        }
        
        // Reserved package names
        if( fqn.startsWith("java.") || fqn.startsWith("javax.") ) {
            String s = Constants.JULIAC_RUNTIME_GENERATED + "." + fqn;
            return s;            
        }
        
        return fqn;
    }

    /**
     * Return the content class name associated to the specified controller and
     * content descriptors. For template components, this method returns the
     * content class name of components created by the template.
     * 
     * @throws IllegalArgumentException
     *      if contentDesc is not a legal content descriptor
     */
    public static String getContentClassName(
		String ctrlDesc, Object contentDesc )
    throws IllegalArgumentException {
        
        /*
         * Adaptation for templates components.
         * Templates are created with something like:
         * Component cTmpl = cf.newFcInstance(
         *      cType, "flatPrimitiveTemplate",
         *      new Object[] { "flatPrimitive", "ClientImpl" });
         */
        if( ctrlDesc.endsWith("Template") ) {
            
            if( !(contentDesc instanceof Object[]) ) {
                String msg =
                    "contentDesc should be an array when "+ctrlDesc+" is used";
                throw new IllegalArgumentException(msg);
            }
            Object[] cont = (Object[]) contentDesc;
            
            if( !(cont[0] instanceof String) ) {
                String msg = "Inner controllerDesc should be a String";
                throw new IllegalArgumentException(msg);
            }
            if( cont[1]!=null && !(cont[1] instanceof String) ) {
                String msg = "Inner contentDesc should be a String";
                throw new IllegalArgumentException(msg);
            }
            
            return (String) cont[1];
        }
        else {
            if( contentDesc!=null && !(contentDesc instanceof String) ) {
                String msg = "contentDesc should be a String";
                throw new IllegalArgumentException(msg);
            }
            return (String) contentDesc;
        }
    }
    
    /**
     * Return the provider prefix (i.e. a slash starting and ending string such
     * as <code>/koch/</code>) inserted in front of the specified controller
     * descriptor.
     * 
     * @param ctrlDesc  the controller descriptor
     * @return  the prefix contained in the specified controller descriptor or
     * 			<code>null</code> if the specified controller descriptor is
     * 			<code>null</code> or if there is no such prefix
     */
    public static String getControllerDescPrefix( String ctrlDesc ) {
        
        if( ctrlDesc == null || ctrlDesc.charAt(0) != '/' ) {
            // No prefix
            return null;
        }
        
        int lastslash = ctrlDesc.lastIndexOf('/');
        if( lastslash!=-1 && lastslash+1 < ctrlDesc.length() ) {
            String prefix = ctrlDesc.substring(0,lastslash+1);
            return prefix;
        }
        
        // No trailing slash character
        return null;
    }
    
    /**
     * Strip the provider prefix (i.e. a slash starting and ending string such
     * as <code>/koch/</code>) inserted in front of the specified controller
     * descriptor. If there is no such prefix, return the controller descriptor
     * unchanged.
     */
    public static String stripControllerDescPrefix( String ctrlDesc ) {
        String prefix = getControllerDescPrefix(ctrlDesc);
        if( prefix == null ) {
            return ctrlDesc;
        }
        String cdesc = ctrlDesc.substring(prefix.length());
        return cdesc;
    }

    /**
     * If <code>fileName</code> denotes an absolute path, return the
     * corresponding file. Else, if <code>fileName</code> denotes a relative
     * path, return the corresponding file rooted from <code>baseDir</code>.
     */
    public static File getFile( File baseDir, String fileName ) {
        File f = new File(fileName);
        if( ! f.isAbsolute() ) {
            f = new File(baseDir,fileName);
        }
        return f;
    }

    /**
     * Return a copy of the specified string, with leading and trailing blank
     * characters (whitespace, tab and newline) omitted.
     */
    public static String trimHeadingAndTrailingBlanks( String src ) {
        
        if( src == null ) {
            return null;
        }
        
        /*
         * i is the index of the first non blank character.
         */
        int i=-1 , len=src.length();
        char c;
        do {
            i++;
            if( i == len ) {
                /*
                 * Either the string is empty or contains only blank characters.
                 */
                return "";
            }
            c = src.charAt(i);
        }
        while( c==' ' || c=='\t' || c=='\n' );
        
        /*
         * j is the index of the first non blank character from the end of src.
         */
        int j = src.length();
        do {
            j--;
            c = src.charAt(j);
        }
        while( c==' ' || c=='\t' || c=='\n' );
    
        String trim = src.substring(i,j+1);
        return trim;
    }

    /**
     * Return the signature of the generic parameter names. Return a string of
     * the form <code>&lt;P1,P2,...&gt;</code>.
     */
    public static String getTypeParameterNamesSignature( String[] tpnames ) {
        
        if( tpnames.length == 0 ) {
            return "";
        }

        StringBuffer sb = new StringBuffer();
        sb.append('<');
        boolean first = true;
        for (String tpname : tpnames) {
            if(first) {first=false;} else {sb.append(',');}
            sb.append(tpname);
        }
        sb.append('>');
        return sb.toString();
    }
    
    /**
     * Return <code>true</code> if the specified file corresponds to a regular
     * class path entry.
     * 
     * Valerio: This method fixes 
     * <a href="http://forge.objectweb.org/tracker/index.php?func=detail&aid=312387&group_id=329&atid=350479">bug #312387</a>.
     * 
     * Christophe M: Files which are not jars, zips, or directories cause {@link
     * NullPointerException} when trying to create class path entries from them.
     * 
     * @param f  the file to check
     * @return   <code>true</code> if the class path entry corresponding to the
     *           specified file is regular
     * 
     * @author Valerio Schiavoni <valerio.schiavoni@gmail.com>
     * @author Christophe Munilla <Christophe.Munilla@inria.fr>
     */
    public static boolean isRegularClassPathEntry( File f ) {
        boolean exits = f.exists();
        boolean isJar = f.getAbsolutePath().endsWith(".jar");
        boolean isZip = f.getAbsolutePath().endsWith(".zip");
        boolean isDirectory = f.isDirectory();        
        return (exits && (isJar || isZip || isDirectory));
    }
    
    /**
     * Return a new component type constructed from the specified array of
     * interface types. Wraps the declared {@link InstantiationExcepton} into a
     * {@link JuliacRuntimeException}.
     * 
     * @param its  the interface types
     * @return     the corresponding component type
     * @throws JuliacRuntimeException  if the component type cannot be created
     * @since 2.5
     */
    public static ComponentType newBasicComponentType( InterfaceType[] its )
    throws JuliacRuntimeException {
        try {
			ComponentType ct = new BasicComponentType(its);
			return ct;
		}
        catch( InstantiationException ie ) {
			throw new JuliacRuntimeException(ie);
		}
    	
    }    
}
