/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.api;

import java.io.IOException;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;

/**
 * Interface for optimization level source code generators.
 * 
 * @param <T>  the type of the structure representing the membrane
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.2
 */
public interface FCSourceCodeGeneratorItf<T> extends JuliacModuleItf {

    /**
     * Test whether the current source code generator accepts the specified
     * controller descriptor.
     * 
     * @param controllerDesc  the controller descriptor
     * @return  <code>true</code> if the current source code generator accepts
     * 			the specified controller descriptor
     * @since 2.2
     */
    public boolean acceptCtrlDesc( Object controllerDesc );
    
    /**
     * Set the prefix for controller descriptors associated with this generator.
     * This prefix is of the form <code>/prefix/</code> and may be specified
     * when several different optimization level source code generators are used
     * jointly to distinguish which optimization level should be used for a
     * particuler component, e.g. <code>/koch/primitive</code> to specify that
     * this is a primitive component that must be handled by the optimization
     * level source code generator associated with the <code>/koch/</code>
     * prefix.
     * 
     * @param prefix  the controller descriptor prefix
     * @since 2.5
     */
    public void setCtrlDescPrefix( String prefix );
    
    /**
     * Generate the source code associated to a Fractal component.
     * The membrane implementation and the initializer implementation are
     * generated.
     * 
     * @param type            the component type
     * @param controllerDesc  the component controller descriptor
     * @param contentDesc     the component content descriptor
     * @return  the initializer class generator for this component
     */
    public InitializerClassGenerator<T> generate(
        Type type, Object controllerDesc, Object contentDesc )
    throws IOException;
    
    /**
     * Generate the source code of the interfaces and interceptors (hence
     * proxies) associated to the specified interface signature. Since
     * interceptors may depend of controller implementations (e.g. lifecycle
     * interceptors use lifecycle controllers), controllers need also to be
     * generated.
     * 
     * @param signature       the interface signature
     * @param controllerDesc  the component controller descriptor
     * @since 2.3.1
     */
    public void generateProxiesImpl( String signature, Object controllerDesc )
    throws IOException;
    
    /**
     * Generate the implementation of the membrane associated to a Fractal
     * component.
     * 
     * @param ct                the component type
     * @param ctrlDesc          the component controller descriptor
     * @param contentClassName  the name of the component content class
     * @return                  data about the membrane implementation
     */
    public MembraneDesc<T> generateMembraneImpl(
        ComponentType ct, String ctrlDesc, String contentClassName )
    throws IOException;

    /**
     * Return the class generator for component interfaces.
     */
    public ProxyClassGeneratorItf getInterfaceClassGenerator( 
        InterfaceType it, MembraneDesc<?> membraneDesc );
}
