/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.api;

import java.util.List;

import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.NoSuchControllerDescriptorException;

/**
 * Interface for retrieving membrane configuration data.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.3
 */
public interface MembraneLoaderItf extends JuliacModuleItf {

	/**
	 * Check that the specified value is a valid membrane definition before
	 * inserting it in the map.
	 * 
	 * @param ctrlDesc  the controller descriptor associated with the membrane
	 * @param cl        the class representing the membrane definition
	 * @return  the class previously associated with the controller descriptor
	 * 			or <code>null</null> if no such controller descriptor was
	 * 			registered before
	 * @throws IllegalArgumentException
	 * 		if the specified membrane definition is not valid
	 * @since 2.5
	 */
	public Class<?> put( String ctrlDesc, Class<?> cl )
	throws IllegalArgumentException;
	
	/**
	 * Return <code>true</code> if the specified membrane descriptor exists.
	 */
	public boolean containsKey( String ctrlDesc );
	
    /**
     * Return the type of the membrane associated with the specified controller
     * descriptor. The component type is needed for parametric membranes. In
     * this case, the attribute controller implementation depends on the
     * attribute control interfaces which is specified in the component type.
     * 
     * @param ct        the component type
     * @param ctrlDesc  the control membrane descriptor
     * @throws NoSuchControllerDescriptorException
     * 		if the specified controller descriptor does not exist
     */
    public InterfaceType[] getMembraneType( ComponentType ct, String ctrlDesc )
    throws NoSuchControllerDescriptorException;

    /**
     * Return the list of controller descriptors associated with the specified
     * control membrane descriptor and component type.
     * 
     * @param ct        the component type
     * @param ctrlDesc  the control membrane descriptor
     * @throws NoSuchControllerDescriptorException
     * 		if the specified controller descriptor does not exist
     */
    public List<ControllerDesc> getCtrlImplLayers(
        ComponentType ct, String ctrlDesc )
    throws NoSuchControllerDescriptorException;

    /**
     * Return the class name of the interceptor class generator associated with
     * the specified controller descriptor and component type. Return
     * <code>null</code> if there is no such interceptor class generator.
     * 
     * @param ct        the component type
     * @param ctrlDesc  the control membrane descriptor
     * @throws NoSuchControllerDescriptorException
     * 		if the specified controller descriptor does not exist
     */
    public String getInterceptorClassGeneratorName(
		ComponentType ct, String ctrlDesc )
    throws NoSuchControllerDescriptorException;

    /**
     * Return the array of interceptor code generator class names associated
     * with the specified controller descriptor and component type. Return
     * <code>null</code> if there is no such interceptor code generator.
     * 
     * @param ct        the component type
     * @param ctrlDesc  the control membrane descriptor
     * @throws NoSuchControllerDescriptorException
     * 		if the specified controller descriptor does not exist
     */
    public String[] getInterceptorSourceCodeGeneratorNames(
		ComponentType ct, String ctrlDesc )
    throws NoSuchControllerDescriptorException;
}
