/***
 * Juliac
 * Copyright (C) 2010-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.conf;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.juliac.Juliac;

/**
 * This class holds data about generators, either source code generators, or
 * class generators, manipulated by {@link JuliacConfig}.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.6
 */
public class GeneratorMap<V> extends HashMap<String,V> {

	private static final long serialVersionUID = 2462164894577262782L;

	protected Juliac jc;
	
	public GeneratorMap( Juliac jc ) {
		this.jc = jc;
	}

	/**
	 * This map stores associations between bytecode generator class names and
	 * their associated source code generator class names.
	 */
	private Map<String,String> names = new HashMap<String,String>();

	/** @since 2.5 */
	@Override
	public GeneratorMap<V> clone() {
		@SuppressWarnings("unchecked")
		GeneratorMap<V> clone = (GeneratorMap<V>) super.clone();
		clone.names = new HashMap<String,String>(names);
		return clone;
	}

    /**
     * Define the mapping between interceptor bytecode and source code
     * generators. The mapping is a colon-separated list of pairs. Each pair
     * contains two values separated by the equal character.
     * 
     * @param def  the mapping between bytecode and source code generators
     * @throws IllegalArgumentException  in case of error in the mapping
     */
    public void set( String def ) throws IllegalArgumentException {
        
        String[] iscgs = def.split(":");
        for (String iscg : iscgs) {
            String[] elts = iscg.split("=");

            if( elts.length != 2 ) {
                String msg = iscg+" should contain only one =";
                throw new IllegalArgumentException(msg);
            }
            
            String asm = elts[0];
            String scg = elts[1];
            set(asm,scg);
        }
    }
    
    public void set( String key, String name ) {
        names.put(key,name);
    }
    
    /** @since 2.5 */
    @Override
    public V get( Object key ) {
    	
    	/*
    	 * key should be typed String, but since key is defined as Object in
    	 * HashMasp, we need to retain Object in order to override get(Object).
    	 * 
    	 * If the source code generator has not been previously registered in
    	 * the map, load and register it (load invokes put.)
    	 */

    	if( containsKey(key) ) {
    		V value = super.get(key);
    		return value;
    	}
    	else {
    		String skey = (String) key;
    		String name = names.containsKey(skey) ? names.get(key) : skey;
    		V value = load(skey,name);
    		return value;
    	}
    }

    /**
     * Load and instantiate the source code generator whose class name is
     * specified and stores the assocation between the specified bytecode
     * generator class name and the instantiated source code generator.  
     * 
     * @param key   the bytecode generator class name
     * @param name  the source code generator class name
     * @return      the instantiated source code generator
     * @since 2.3
     */
    private V load( String key, String name ) {
    	
    	/*
    	 * Do not record the association in #names. The rationale for #names is
    	 * to store associations up until classes can be loaded. If the class
    	 * can be loaded directly as this is the case here in method
    	 * load(String,String), it is not necessary to use #names. By the way,
    	 * this would lead to ConcurrentModificationException when
    	 * load(String,String) is invoked from load().
    	 */
    	JuliacConfig jconf = jc.getJuliacConfig();
        Class<V> cl = jconf.load(name);
        V iscg = jconf.instantiate(cl);
        put(key,iscg);
        return iscg;
    }
}
