/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.conf;

import java.util.Arrays;
import java.util.logging.Logger;

/**
 * The source code and bytecode compatibility levels.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.2
 */
public enum JDKLevel {
    
    JDK1_1("1.1"), JDK1_2("1.2"), JDK1_3("1.3"), JDK1_4("1.4"), JDK1_5("1.5"),
    JDK1_6("1.6"), JDK1_7("1.7");
    
    /**
     * @since 2.2.3
     */
    private JDKLevel( String version ) {
        this.version = version;
    }

    /**
     * @since 2.2.3
     */
    private String version;
    
    /**
     * @since 2.2.3
     */
    public String getVersion() {
        return version;
    }
    
    /**
     * Return the default compatibility level specified by the
     * <code>java.specification.version</code> system property or {@link
     * #JDK1_5} if the system property specifies a level which can not be parsed
     * by {@link #valueOfIAE(String)}.
     * 
     * @param logger
     * 		the logger for reporting a warning message if the default
     * 		compatibility level can not be retrieved from the
     * 		<code>java.specification.version</code> system property or
     * 		<code>null</code> if reporting the warning message is not necessary
     * @since 2.2.6
     */
    public static JDKLevel getDefaultLevel( Logger logger ) {
    	try {
        	String v = System.getProperty("java.specification.version");
        	JDKLevel level = valueOfIAE(v);
        	return level;
    	}
    	catch( IllegalArgumentException iae ) {
    		if( logger != null ) {
    			String msg =
    				iae.getMessage()+
    				". This value is returned by the "+
    				"java.specification.version system property. "+
    				"1.5 will be assumed instead.";
    			logger.warning(msg);
    		}
            return JDK1_5;
    	}
    }

    /**
     * Return the default compatibility level for Java source code.
     * 
     * @param logger
     * 		the logger for reporting a warning message if the default
     * 		compatibility level can not be retrieved from the
     * 		<code>java.specification.version</code> system property or
     * 		<code>null</code> if reporting the warning message is not necessary
     * @since 2.2.3
     */
    public static JDKLevel getDefaultSourceLevel( Logger logger ) {
    	JDKLevel level = getDefaultLevel(logger);
    	return level;
    }
    
    /**
     * Return the default compatibility level for Java bytecode.
     * 
     * @param logger
     * 		the logger for reporting a warning message if the default
     * 		compatibility level can not be retrieved from the
     * 		<code>java.specification.version</code> system property or
     * 		<code>null</code> if reporting the warning message is not necessary
     * @since 2.2.3
     */
    public static JDKLevel getDefaultTargetLevel( Logger logger ) {
    	JDKLevel level = getDefaultLevel(logger);
    	return level;
    }
 
    /**
     * <p>
     * Return the constant corresponding to the specified compatibility version
     * number.
     * </p>
     * 
     * <p>
     * This method has been defined even though method <code>valueOf</code>
     * exists for enum values, for the following two reasons:
     * </p>
     * 
     * <ul>
     * <li>we want to retrieve the constant corresponding to the version number
     *     (e.g. 1.4, 1.6, etc.), not the one corresponding to the constant
     *     identifier (e.g. JDK1_4, JDK1_6, etc.),</li>
     * <li>we want to be able to throw an {@link IllegalArgumentException} with
     *     a custom error message in case of illegal value.</li>
     * </ul>
     * 
     * @param v  the compatibility version number
     * @return   the corresponding {@link JDKLevel} constant
     * @throws IllegalArgumentException
     *      if <code>v</code> is not a legal compatibility version number
     * @since 2.2.3
     */
    public static JDKLevel valueOfIAE( String v )
    throws IllegalArgumentException {
        
        for (JDKLevel level : JDKLevel.values()) {
            if( level.version.equals(v) ) {
                return level;
            }
        }
        
        String[] versions = getVersions();
        String s = Arrays.toString(versions);
        String msg =
        	"Java compatibility version number "+v+
        	" is not among the supported ones: "+s;
        throw new IllegalArgumentException(msg);
    }

    /**
     * Return all version numbers defined by {@link JDKLevel} constants.
     * 
     * @since 2.2.3
     */
    public static String[] getVersions() {
        JDKLevel[] levels = JDKLevel.values();
        String[] versions = new String[ levels.length ];
        for (int i = 0; i < levels.length; i++) {
            versions[i] = levels[i].getVersion();
        }
        return versions;
    }
}
