/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.conf;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.juliac.CompilationRound;
import org.objectweb.fractal.juliac.CompilationRounds;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.JuliacModuleItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.RuntimeClassNotFoundException;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.LifeCycleSourceCodeGenerator;

/**
 * This class manages the Juliac configuration parameters.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.1.4
 */
public class JuliacConfig {
    
	private Juliac jc;
	
    public JuliacConfig( Juliac jc ) {
        
    	this.jc = jc;
    	iclgs = new GeneratorMap<InterceptorClassGenerator>(jc);
        iscgs = new InterceptorSourceCodeGeneratorMap(jc);
        
		// Default class generator
        iclgs.put(
    		Constants.JULIA_INTERCEPTOR_CLASS_GENERATOR,
            new InterceptorClassGenerator() );

        // Default source code generator
		iscgs.put(
            Constants.JULIA_LIFECYCLE_CODE_GENERATOR,
            new LifeCycleSourceCodeGenerator() );	
    }
    
    
    // ----------------------------------------------------------------------
    // Juliac properties
    // ----------------------------------------------------------------------
    
    /**
     * The base directory for locations specified in the parameters.
     */
    private File baseDir = new File(".");
    
    /**
     * The list of locations containing source files. A location denotes either
     * an absolute path or a path relative to {@link #baseDir}, for a directory
     * or a jar file.
     */
    private List<String> srcs = new ArrayList<String>();
    
    /**
     * The list of locations containing source files for librairies. A location
     * denotes either an absolute path or a path relative to {@link #baseDir},
     * for a directory or a jar file.
     */
    private List<String> srclibs = new ArrayList<String>();
    
    /**
     * The root package for generated code.
     */
    private String pkgRoot = "";

    /**
     * The name of the directory where generated files are stored. The name
     * denotes either an absolute path or a path relative to {@link #baseDir}.
     * The resulting {@link File} is stored in {@link #genDirName}.
     */
    private String genDirName = "target/generated-sources/juliac";
    
    /**
     * The directory where generated files are stored.
     */
    private File genDir;
    
    /**
     * The name of the directory where generated files are compiled. The name
     * denotes either an absolute path or a path relative to {@link #baseDir}.
     * The resulting {@link File} is stored in {@link #classDir}.
     */
    private String classDirName = "target/classes";
    
    /**
     * The directory where generated files are compiled.
     */
    private File classDir;
    
    /**
     * The name of the jar file where compiled generated files are packaged. The
     * resulting {@link File} is stored in {@link #classJarFile}.
     * 
     * @since 2.2.4
     */
    private String classJarFileName = "target/juliac.jar";
    
    /**
     * The jar file where compiled generated files are packaged.
     * @since 2.2.4
     */
    private File classJarFile;
    
    /**
     * JDK level for source code.
     * @since 2.2.2
     */
    private JDKLevel sourceLevel = JDKLevel.getDefaultSourceLevel(getLogger());
    
    /**
     * JDK level for class files.
     * @since 2.2.2
     */
    private JDKLevel targetLevel = JDKLevel.getDefaultTargetLevel(getLogger());
    
    /**
     * The class generators for component interceptors.
     * 
     * @since 2.2.6
     */
    private GeneratorMap<InterceptorClassGenerator> iclgs;
    
    /**
     * The source code generators for component interceptors.
     * 
     * @since 2.2.6
     */
    private InterceptorSourceCodeGeneratorMap iscgs;
    
    private ClassLoader classLoader;
    private CompilationRounds rounds;

	/**
	 * The default logger.
	 * @since 2.2.4
	 */
	private Logger logger = null;

	/**
	 * The name of the default logger.
	 * @since 2.2.4
	 */
	private final static String LOGGER_NAME = "org.objectweb.fractal.juliac";
    
	/**
	 * The logger for the compiler plugin.
	 * @since 2.2.4
	 */
	private Logger compilationLogger = null;

	/**
	 * The name of the logger for the compiler plugin.
	 * @since 2.2.4
	 */
	private final static String COMPILATION_LOGGER_NAME =
		"org.objectweb.fractal.juliac.plugin.compiler";

	/**
	 * <code>true</code> if compilation warnings should be reported.
	 * @since 2.2.4
	 */
	private boolean compilationWarnings = false;
    
	
    // ----------------------------------------------------------------------
    // Setter/getter methods for Juliac file and directory related properties
    // ----------------------------------------------------------------------
    
    public void setPkgRoot( String pkgRoot ) { this.pkgRoot = pkgRoot; }
    public String getPkgRoot() { return pkgRoot; }
    
    public void setBaseDir( File baseDir ) { this.baseDir = baseDir; }
    public File getBaseDir() { return baseDir; }
    
    public String[] getSrcs() {
        /*
         * Return a copy in order to prevent unwanted adding in the list.
         * Elements must be added in the list by invoking either {@link
         * #addSrc(String)} or {@link #addSrcs(String[])}.
         */
        return srcs.toArray(new String[srcs.size()]);
    }
    
    /**
     * Add the specified location to the list of locations. A location denotes
     * either an absolute path or a path relative to {@link #baseDir}, for a
     * directory or a jar file. All source files contained in the specified
     * location are registered with the {@link CompilationRounds} API for being
     * compiled when requested (see {@link #compile()}) or processed by the
     * juliac-spoon module (see {@link SpoonSupportItf}.)
     * 
     * @param src  the location to add
     * @throws IOException
     *      if the location does not exist or an error occurred while fetching
     *      the source files contained in the location
     */
    public void addSrc( String src ) throws IOException {
        
        /*
         * Retrieve all source files contained in src.
         */
        File baseDir = getBaseDir();
        List<SourceFile> sfs = new ArrayList<SourceFile>();
        SourceFile.addAllJavaFiles(baseDir,src,sfs);
        
        /*
         * Register the source files with the current compilation round.
         */
        CompilationRounds rounds = getCompilationRounds();
        CompilationRound round = rounds.getCurrentCompilationRound();
        round.addInput(sfs);

        srcs.add(src);
    }
    
    /**
     * Invoke {@link #addSrc(String)} for each element contained in the
     * specified array of locations.
     */
    public void addSrcs( String[] srcs ) throws IOException {
        for (String src : srcs) {
            addSrc(src);
        }
    }
    
    public String[] getSrclibs() {
        /*
         * Return a copy in order to prevent unwanted adding in the list.
         * Elements must be added in the list by invoking either {@link
         * #addSrclib(String)} or {@link #addSrclibs(String[])}.
         */
        return srclibs.toArray(new String[srclibs.size()]);
    }
    
    /**
     * Add the specified location to the list of library locations. A location
     * denotes either an absolute path or a path relative to {@link #baseDir},
     * for a directory or a jar file.
     * 
     * @param src  the location to add
     */
    public void addSrclib( String srclib ) {
        srclibs.add(srclib);
    }
    
    /**
     * Invoke {@link #addSrclibs(String)} for each element contained in the
     * specified array of locations.
     */
    public void addSrclibs( String[] srclibs ) {
        for (String srclib : srclibs) {
            addSrclib(srclib);
        }
    }
    
    public File getGenDir() throws IOException {
        if( genDir == null ) {
            File baseDir = getBaseDir();
            String genDirName = getGenDirName();
            genDir = initDir(baseDir,genDirName);            
        }
        return genDir;
    }
    public String getGenDirName() { return genDirName; }
    public void setGenDirName( String genDirName ) throws IOException {
        File baseDir = getBaseDir();
        genDir = initDir(baseDir,genDirName);
    }
    
    public File getClassDir() throws IOException {
        if( classDir == null ) {
            File baseDir = getBaseDir();
            String classDirName = getClassDirName();
            classDir = initDir(baseDir,classDirName);
        }
        return classDir;
    }
    public String getClassDirName() { return classDirName; }
    public void setClassDirName( String classDirName ) throws IOException {
        File baseDir = getBaseDir();
        classDir = initDir(baseDir,classDirName);
    }
    
    public File getClassJarFile() {
    	if( classJarFile == null ) {
    		String classJarFileName = getClassJarFileName();
    		classJarFile = new File(baseDir,classJarFileName);
    	}
    	return classJarFile;
    }
    public String getClassJarFileName() { return classJarFileName; }
    public void setClassJarFileName( String classJarFileName ) {
        File baseDir = getBaseDir();
        classJarFile = new File(baseDir,classJarFileName);
    }

    
    // ----------------------------------------------------------------------
    // Setter/getter methods for Java compilation related properties
    // ----------------------------------------------------------------------
    
    public JDKLevel getSourceLevel() { return sourceLevel; }
    public void setSourceLevel( JDKLevel sourceLevel ) {
        this.sourceLevel = sourceLevel;
    }

    public JDKLevel getTargetLevel() { return targetLevel; }
    public void setTargetLevel( JDKLevel targetLevel ) {
        this.targetLevel = targetLevel;
    }
    
    
    // ----------------------------------------------------------------------
    // Getter methods for interceptor source code and class generators
    // ----------------------------------------------------------------------

    /**
     * @since 2.2.6
     */
    public GeneratorMap<InterceptorClassGenerator> interceptorClassGenerators() {
    	return iclgs;
    }

    /**
     * @since 2.2.6
     */
    public InterceptorSourceCodeGeneratorMap interceptorSourceCodeGenerators() {
    	return iscgs;
    }
    
    // ----------------------------------------------------------------------
    // Setter/getter methods for loaders and compilation rounds
    // ----------------------------------------------------------------------
    
    public CompilationRounds getCompilationRounds() {
        if( rounds == null ) {
            rounds = new CompilationRounds(this);
        }
        return rounds;
    }

    public ClassLoader getClassLoader() {
        if( classLoader == null ) {
            classLoader = getClass().getClassLoader();
        }
        return classLoader;
    }
    
    public void setClassLoader( ClassLoader classLoader ) {
        this.classLoader = classLoader;
    }

    /**
     * Load the class whose name is specified.
     * 
     * @param <T>   the type of the class to load
     * @param name  the name of the class to load
     * @return      the class
     * @throw RuntimeClassNotFoundException  if the class cannot be loaded
     */
    public <T> Class<T> load( String name )
	throws RuntimeClassNotFoundException {
        
        ClassLoader classloader = getClassLoader();
        Class<?> c = null;
        try {
			c = classloader.loadClass(name);
		}
        catch (ClassNotFoundException e) {
            throw new RuntimeClassNotFoundException(name);
		}
        
        @SuppressWarnings("unchecked")
        Class<T> cl = (Class<T>) c;
        return cl;
    }

    /**
     * Instantiate the specified class by calling the default constructor.
     * Wrap exceptions thrown by {@link Class#newInstance()} ({@link
     * InstantiationException} and {@link IllegalAccessException}) into {@link
     * JuliacRuntimeException}.
     * 
     * @param <T>        the type of the class to instantiate
     * @param className  the name of the class to instantiate
     * @return           the instance
     * @throws JuliacRuntimeException  in case of error during the instantiation
     */
    public <T> T instantiate( Class<T> cl ) throws JuliacRuntimeException {
        
        try {
            T o = cl.newInstance();
            return o;
        }
        catch (InstantiationException e) {
            throw new JuliacRuntimeException(e);
        }
        catch (IllegalAccessException e) {
            throw new JuliacRuntimeException(e);
        }
    }
    

    // ----------------------------------------------------------------------
    // Logging facilities
    // ----------------------------------------------------------------------

    /**
     * Return the default logger.
     * 
     * @since 2.2.4
     */
	public Logger getLogger() {
	    
	    if( logger == null ) {
	        logger = Logger.getLogger(LOGGER_NAME);
	        Logger parent = logger.getParent();
	        Handler[] handlers = parent.getHandlers();
	        for (Handler handler : handlers) {
	            Formatter newFormatter = new LoggerFormatter();
	            handler.setFormatter(newFormatter);
	        }
	    }
	
	    return logger;
	}

    /**
     * Return the logger for the compiler plugin.
     * 
     * @since 2.2.4
     */
	public Logger getCompilationLogger() {
	    
	    if( compilationLogger == null ) {
	    	compilationLogger = Logger.getLogger(COMPILATION_LOGGER_NAME);
	        Logger parent = compilationLogger.getParent();
	        Handler[] handlers = parent.getHandlers();
	        for (Handler handler : handlers) {
	            Formatter newFormatter = new LoggerFormatter();
	            handler.setFormatter(newFormatter);
	        }
	        
	        // By default, the compilation loggers only logs errors
	        setLevel(compilationLogger,Level.SEVERE);
	    }
	
	    return compilationLogger;
	}

	/**
	 * Define a log level.
	 * 
	 * @param logger  the logger
	 * @param level   the log level
	 * @since 2.2.4
	 */
	public static void setLevel( Logger logger, Level level ) {
		logger.setLevel(level);
        Logger parent = logger.getParent();
        Handler[] handlers = parent.getHandlers();
        for (Handler handler : handlers) {
            handler.setLevel(level);
        }
	}
	
	/**
	 * @since 2.2.4
	 */
	public void setCompilationWarnings( boolean compilationWarnings ) {
		this.compilationWarnings = compilationWarnings;
	}

	/**
	 * @since 2.2.4
	 */
	public boolean getCompilationWarnings() {
		return compilationWarnings;
	}

	
	// ----------------------------------------------------------------------
    // Modules and services management
    // ----------------------------------------------------------------------

	private Map<Class<? extends JuliacModuleItf>,List<JuliacModuleItf>> services =
		new HashMap<Class<? extends JuliacModuleItf>,List<JuliacModuleItf>>();

	/**
	 * Load and initialize a module.
	 * 
	 * @param name  the fully-qualified name of the module or a constant value
	 * 				defined in {@link Module}
	 * @return      the service exported by the module
	 * @throws RuntimeClassNotFoundException  if the module cannot be loaded
	 * @since 2.4
	 */
	public <T> T loadModule( String name )
	throws IOException, JuliacRuntimeException {
		
    	String className = null;
    	try {
        	Module m = Module.valueOf(name);
        	className = m.getClassName();
    	}
    	catch( IllegalArgumentException iae ) {
    		className = name;
    	}

    	/*
		 * Method init(Juliac) is supposed to register service implementations
		 * by calling register(Class<T>,T).
		 */
		Class<JuliacModuleItf> cl = load(className);
		JuliacModuleItf service = instantiate(cl);
		service.init(jc);
		
		@SuppressWarnings("unchecked")
		T t = (T) service;
		return t;
	}
	
	/**
	 * Load and initialize a module if the service exported by the module has
	 * not been registered before.
	 * 
	 * @param name  the fully-qualified name of the module or a constant value
	 * 				defined in {@link Module}
	 * @param itf   the service type
	 * @return      the registered service or the service exported by the module
	 * @throws RuntimeClassNotFoundException  if the module cannot be loaded
	 * @since 2.5
	 */
	public <T extends JuliacModuleItf>
		T loadModuleIfNew( String name, Class<T> itf )
	throws IOException, JuliacRuntimeException {

		try {
			T t = lookupUnique(itf);
			return t;
		}
		catch( JuliacRuntimeException jre ) {
			// cast to avoid bound compilation issue under javac
			@SuppressWarnings("unchecked")
			T t = (T) loadModule(name);
			return t;
		}
	}
	
	/**
	 * Unload modules with registered services.
	 * 
	 * @since 2.4
	 */
	public void unloadModules() throws IOException {
		
		/*
		 * Method close() is supposed to unregister service implementations by
		 * calling unregister(Class<T>,T).
		 */
		Collection<List<JuliacModuleItf>> services =
			new ArrayList<List<JuliacModuleItf>>(this.services.values());
		
		for (List<JuliacModuleItf> e : services) {
			List<JuliacModuleItf> copy = new ArrayList<JuliacModuleItf>(e);
			for (JuliacModuleItf service : copy) {
				service.close(jc);
			}
		}
	}
	
	/**
	 * Register a service implementation.
	 * 
	 * @param itf  	the service type
	 * @param impl  the service implementation
	 * @since 2.4
	 */
	public <T extends JuliacModuleItf> void register( Class<T> itf, T impl ) {
		List<JuliacModuleItf> services = null;
		if( this.services.containsKey(itf) ) {
			services = this.services.get(itf);
		}
		else {
			services = new ArrayList<JuliacModuleItf>();
			this.services.put(itf,services);
		}
		services.add(impl);
	}
	
	/**
	 * Retrieve service implementations.
	 * 
	 * @param itf  the service type
	 * @return     the registered service implementations
	 * @since 2.4
	 */
	public <T extends JuliacModuleItf> List<T> lookup( Class<T> itf ) {
		
		List<JuliacModuleItf> services =
			this.services.containsKey(itf) ?
			this.services.get(itf) :
			new ArrayList<JuliacModuleItf>();
		
		@SuppressWarnings("unchecked")
		List<T> t = (List<T>) services;
		return t;
	}
	
	/**
	 * Retrieve a service implementation.
	 * 
	 * @param itf  the service type
	 * @return     the registered service implementation
	 * @throws JuliacRuntimeException
	 * 		if more than one service or zero service implementations are
	 * 		registered for the specified service type
	 * @since 2.4
	 */
	public <T extends JuliacModuleItf> T lookupUnique( Class<T> itf )
	throws JuliacRuntimeException {
		List<T> services = lookup(itf);
		if( services.size() == 1 ) {
			return services.get(0);
		}
		else {
			String msg =
				"No unique registered service type: "+itf.getName()+
				". Got instead: "+services.size();
			throw new JuliacRuntimeException(msg);
		}
	}
	
	/**
	 * Unregister a service implementation.
	 * 
	 * @param itf   the service type
	 * @param impl  the service implementation
	 * @since 2.4
	 */
	public <T extends JuliacModuleItf> void unregister( Class<T> itf, T impl ) {
		if( services.containsKey(itf) ) {
			List<JuliacModuleItf> services = this.services.get(itf);
			services.remove(impl);
		}
	}
	
	
	// ----------------------------------------------------------------------
    // Implementation specific
    // ----------------------------------------------------------------------

	/**
     * Check that the specified file name corresponds to an existing directory.
     * If not, create the corresponding hierarchy of directories.
     * 
     * @param baseDir  the base directory
     * @param dirName  the file name
     * @return  the bottom-most created directory
     * 
     * @throws IOException  if the file name exists but is not a directory
     */
    private static File initDir( File baseDir, String dirName )
    throws IOException {
        
        File dir = Utils.getFile(baseDir,dirName);
        if( dir.exists() && !dir.isDirectory() ) {
            String msg = dir.getAbsolutePath()+" is not a directory";
            throw new IOException(msg);
        }
        if( ! dir.exists() ) {
            dir.mkdirs();
        }
        return dir;
    }
}
