/***
 * Juliac
 * Copyright (C) 2011-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.conf;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.extensions.Membrane;
import org.objectweb.fractal.fraclet.types.Constants;

/**
 * Helper class for dealing with @{@link Membrane}-annotated classes.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.3.1
 */
public class MembraneHelper {

	/**
	 * Check that the class associated with the specified
	 * 		@{@link Membrane}<code>(controllerDesc=...)</code>
	 * annotation references a membrane definition, i.e. a class that is
	 * annotated with
	 * 		@{@link Membrane}<code>(desc="...")</code>.
	 * 
	 * @param jm    the {@link Membrane} annotation
	 * @param name  the name of the class annotated with {@link Membrane}
	 * @return      the membrane definition
	 * @throws IllegalArgumentException
	 * 		if the referenced class is not a membrane definition
	 */
	public static Class<?> getMembraneDef( Membrane jm, String name ) {
		Class<?> cl = jm.controllerDesc();
		Membrane mdef = cl.getAnnotation(Membrane.class);
		if( mdef == null ) {
			String msg =
				"The "+cl.getName()+" membrane class specified in the "+
				"@Membrane(controllerDesc=...) annotation on the "+name+
				" class should be annotated with @Membrane(desc=\"...\").";
			throw new IllegalArgumentException(msg);
		}
		return cl;
	}
	
	/**
	 * Check that the specified membrane definition is consistent.
	 */
	public static void check( Class<?> cl ) throws IllegalArgumentException {

		// Check that the class is annotated with @Membrane
		Membrane mdef = cl.getAnnotation(Membrane.class);
		if( mdef == null ) {
			String msg = "No @Membrane annotation on class "+cl.getName();
			throw new IllegalArgumentException(msg);
		}
		
		/*
		 * Check that either the controller, template, controllerDesc,
		 * templateDesc parameters are set or the desc, generator, interceptors
		 * parameters are set.
		 */
		boolean isControllerSet = ! mdef.controller().equals(Constants.EMPTY);
		boolean isTemplateSet = ! mdef.template().equals(Constants.EMPTY);
		boolean isControllerDescSet = ! mdef.controllerDesc().equals(Constants.class);
		boolean isTemplateDescSet = ! mdef.templateDesc().equals(Constants.class);
		boolean isDescSet = ! mdef.desc().equals(Constants.EMPTY);
		boolean isGeneratorSet = ! mdef.generator().equals(Constants.class);
		boolean isInterceptorsSet = mdef.interceptors().length != 0;
		
		boolean first =
			isControllerSet | isTemplateSet | isControllerDescSet |
			isTemplateDescSet;
		boolean second = isDescSet | isGeneratorSet | isInterceptorsSet;
		if( first && second ) {
			String msg =
				"{controller,template,controllerDesc,templateDesc} on one side"+
				" and {desc,generator,interceptors} on the other side "+
				"cannot be set jointly in "+mdef+" for class "+cl.getName();
			throw new IllegalArgumentException(msg);
		}
		if( !first && !second ) {
			String msg =
				"Either {controller,template,controllerDesc,templateDesc} "+
				"or {desc,generator,interceptors} should be set in "+
				mdef+" for class "+cl.getName();
			throw new IllegalArgumentException(msg);
		}
		
		/*
		 * Check that the inheritance hierarchy contains at least one @Membrane
		 * annotated class.
		 */
		List<Class<?>> classes = new ArrayList<Class<?>>();
		addSuperClasses(cl,classes);
		boolean membraneAnnotated = false;
		for (Class<?> c : classes) {
			Membrane m = c.getAnnotation(Membrane.class);
			if( m != null ) {
				membraneAnnotated = true;
				break;
			}
		}
		if( ! membraneAnnotated ) {
			String msg =
				"No @Membrane annotated class in the inheritance hierarchy of "+
				cl.getName();
			throw new IllegalArgumentException(msg);
		}
	}
	
	/**
	 * Return the generator class associated with the specified membrane
	 * definition. Return <code>null</code> if no generator class is defined.
	 */
	public static Class<?> getGenerator( Class<?> cl )
	throws IllegalStateException {

		List<Class<?>> classes = new ArrayList<Class<?>>();
		addSuperClasses(cl,classes);

		for (Class<?> c : classes) {
			Membrane mdef = c.getAnnotation(Membrane.class);
			if( mdef != null ) {
				Class<?> generator = mdef.generator();
				if( ! generator.equals(Constants.class) ) {
					return generator;
				}
			}
		}

		// No generator specified
		return null;
	}

	/**
	 * Return the interceptor classes associated with the specified membrane
	 * definition. Return <code>null</code> if no interceptor class is defined.
	 */
	public static Class<?>[] getInterceptors( Class<?> cl )
	throws IllegalStateException {
		
		List<Class<?>> classes = new ArrayList<Class<?>>();
		addSuperClasses(cl,classes);

		for (Class<?> c : classes) {
			Membrane mdef = c.getAnnotation(Membrane.class);
			if( mdef != null ) {
				Class<?>[] interceptors = mdef.interceptors();
				if( interceptors.length != 0 ) {
					return interceptors;
				}
			}
		}
		
		// No interceptor class specified
		return null;
	}
	
	/**
	 * Add the super membrane definitions of the specified membrane definition
	 * to the specified list. Recurse into the inheritance hierarchy.
	 */
	public static void addSuperClasses( Class<?> cl, List<Class<?>> classes ) {
		
		// Stop the recursion when java.lang.Object is reached
		if( cl.equals(Object.class) ) {
			return;
		}
		classes.add(cl);

		// Super class
		Class<?> sup = cl.getSuperclass();
		addSuperClasses(sup,classes);
	}
	
	/**
	 * Return the list of {@link Controller}-annotated fields declared in the
	 * inheritance hierarchy of the specified class.
	 */
	public static List<Field> getControllerAnnotatedFields( Class<?> cl ) {

		List<Class<?>> classes = new ArrayList<Class<?>>();
		MembraneHelper.addSuperClasses(cl,classes);

		List<Field> fields = new ArrayList<Field>();
		Set<String> names = new HashSet<String>();
		
		for (Class<?> c : classes) {
			Field[] fs = c.getDeclaredFields();
			for (Field f : fs) {
				
				// Skip fields with no @Controller annotation
				if( ! f.isAnnotationPresent(Controller.class) ) {
					continue;
				}
				
				String name = f.getName();
				/*
				 * Do not add the field if another one with the same name has
				 * already been declared previously in the inheritance
				 * hierarchy.
				 */
				if( ! names.contains(name) ) {
					fields.add(f);
					names.add(name);
				}
			}
		}
		
		return fields;
	}
}
