/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.desc;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.opt.AbstractComponentFactoryClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.util.Fractal;

/**
 * This class generates the source code of a component {@link
 * org.objectweb.fractal.api.factory.Factory}. This source code generator visits
 * a model of an ADL file.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.1.6
 */
public class ComponentDescFactoryClassGenerator
extends AbstractComponentFactoryClassGenerator {
    
	private Juliac jc;
    private ComponentDesc cdesc;
    
    public ComponentDescFactoryClassGenerator(
		Juliac jc, ComponentDesc cdesc, String targetname ) {
        
        super(targetname);
        this.jc = jc;
        this.cdesc = cdesc;
    }
    
    /**
     * Generate the <code>newFcInstance999</code> methods.
     * 
     * This method is invoked from
     * InitializerUltraCompCtrlClassGenerator#generateNFICMInstantiateMembrane
     * and is thus declared as public.
     * 
     * @param pw     the print writer where the factory is to be outputted
     * @param declareLocalVars
     *      <code>true</code> if the variables which hold references to
     *      components need to be declared
     */
    public CompDesc generateNewFcInstanceMethods(
		PrintWriter pw, boolean declareLocalVars )
    throws IOException {
        
        /*
         * Generate the code of the factory.
         */
        generate(pw,declareLocalVars);
        
        /*
         * Traverse the architecture to retrieve component identifiers.
         */
        List<String> compIDs = new ArrayList<String>();
        addCompID(cdesc,compIDs);
        
        String id = cdesc.getID();
        ComponentType ct = cdesc.getCT();
        CompDesc cd = new CompDesc(id,ct,compIDs);
        return cd;
    }
    
    /**
     * Add the identifier of the specified component and of all its
     * subcomponents to the specified list.
     * 
     * @param cdesc    the component to traverse
     * @param compIDs  the list of identifiers
     */
    private static void addCompID( ComponentDesc cdesc, List<String> compIDs ) {
        
        String id = cdesc.getID();
        compIDs.add(id);
        
        List<ComponentDesc> subcompdescs = cdesc.getSubComponents();
        for (ComponentDesc subcompdesc : subcompdescs) {
            addCompID(subcompdesc,compIDs);
        }
    }

    
    // ------------------------------------------------------------------
    // Component factory code generation
    // ------------------------------------------------------------------

    /**
     * Generate a factory for the current component.
     * 
     * @param pw     the print writer where the factory is to be outputted
     * @param declareLocalVars
     *      <code>true</code> if the variables which hold references to
     *      components need to be declared
     */
    protected void generate( PrintWriter pw, boolean declareLocalVars )
    throws IOException {
        generateInstantiations(cdesc,pw,declareLocalVars);
        generateHierarchy(cdesc,pw);
        generateBindings(cdesc,pw);
    }
    
    /*
     * In the following methods, cdesc is added systematicaly as a parameter,
     * although cdesc is an instance variable. We need this extra parameter as
     * these methods may be called recursively with different cdescs.
     */
    
    protected void generateInstantiations(
    	ComponentDesc cdesc, PrintWriter pw, boolean declareLocalVars )
	throws IOException {
        
    	String ctrlDesc = cdesc.getCtrlDesc();
    	ComponentType ct = cdesc.getCT();
    	String contentClassName = cdesc.getContentClassName();
    	String id = cdesc.getID();
    	
        /*
         * Generate the membrane code for the component.
         */
        FCSourceCodeGeneratorItf<?> fcscg =
        	jc.getFCSourceCodeGenerator(ctrlDesc);
        ClassGeneratorItf icg = fcscg.generate(ct,ctrlDesc,contentClassName);
        
        /*
         * Instantiate the component.
         */
        String initializerClassName = icg.getTargetClassName();
        
        if( declareLocalVars ) {
            pw.print  (Component.class.getName());
            pw.print  (" ");
        }
        pw.print  (id);
        pw.print  (" = new ");
        pw.print  (initializerClassName);
        pw.println("().newFcInstance();");
        
        generateComponentName(cdesc,pw);
        generateAttributes(cdesc,pw);

        /*
         * Iterate on sub-components.
         */
        List<ComponentDesc> subComponents = cdesc.getSubComponents();
        for (ComponentDesc sub : subComponents) {
            generateInstantiations(sub,pw,declareLocalVars);
        }
    }
    
    protected void generateComponentName( ComponentDesc cdesc, PrintWriter pw ) {
        
    	String id = cdesc.getID();
    	String name = cdesc.getName();
    	
        pw.print  (Fractal.class.getName());
        pw.print  (".getNameController(");
        pw.print  (id);
        pw.print  (").setFcName(\"");
        pw.print  (name);
        pw.println("\");");
    }
    
    protected void generateAttributes( ComponentDesc cdesc, PrintWriter pw ) {

    	Set<String> names = cdesc.getAttributeNames();
    	String id = cdesc.getID();

    	for( String name : names ) {
            
            AttributeDesc adesc = cdesc.getAttribute(name);
            String attributeController = adesc.getAttributeController();
            String mname = name.substring(0,1).toUpperCase()+name.substring(1);
            String settername = "set"+mname;
            String value = adesc.getValue();
            
            pw.print  ("((");
            pw.print  (attributeController);
            pw.print  (")");
            pw.print  (Fractal.class.getName());
            pw.print  (".getAttributeController(");
            pw.print  (id);
            pw.print  (")).");
            pw.print  (settername);
            
            /*
             * Dependending on the type of the attribute, determine the
             * delimiter for its value: " for strings, ' for characters, nothing
             * in other cases.
             */
            UnifiedClass atype = getAttrType(attributeController,settername);
            String delimiter = atype.getValueDelimiter();

			pw.print  ("(");
            pw.print  (delimiter);
            pw.print  (value);
            pw.print  (delimiter);
            pw.println(");");
        }
    }
    
    protected void generateHierarchy( ComponentDesc cdesc, PrintWriter pw ) {
    	
    	List<ComponentDesc> subComponents = cdesc.getSubComponents();
    	String id = cdesc.getID();
    	
        for (ComponentDesc sub : subComponents) {
            String subid = sub.getID();
            pw.print  (Fractal.class.getName());
            pw.print  (".getContentController(");
            pw.print  (id);
            pw.print  (").addFcSubComponent(");
            pw.print  (subid);
            pw.println(");");
        }
        
        /*
         * Iterate on sub-components.
         */
        for (ComponentDesc sub : subComponents) {
            generateHierarchy(sub,pw);
        }
    }

    protected void generateBindings( ComponentDesc cdesc, PrintWriter pw ) {
    	
    	List<BindingDesc> bdescs = cdesc.getBindingDescs();
    	String id = cdesc.getID();
    	List<ComponentDesc> subComponents = cdesc.getSubComponents();

    	for (BindingDesc bdesc : bdescs) {
            
            BindingType btype = bdesc.getBindingType();
            String cltItfName = bdesc.getCltItfName();
            String srvItfName = bdesc.getSrvItfName();
            ComponentDesc srv = bdesc.getSrv();
            String srvid = srv.getID();
            
            pw.print  (Fractal.class.getName());
            pw.print  (".getBindingController(");
            pw.print  (id);
            pw.print  (").bindFc(\"");
            pw.print  (cltItfName);
            pw.print  ("\",");
            if( btype.equals(BindingType.IMPORT) ) {
                pw.print  (Fractal.class.getName());
                pw.print  (".getContentController(");
                pw.print  (srvid);
                pw.print  (").getFcInternalInterface(\"");
            }
            else {
                pw.print  (srvid);
                pw.print  (".getFcInterface(\"");
            }
            pw.print  (srvItfName);
            pw.println("\"));");
        }
        
        /*
         * Iterate on sub-components.
         */
        for (ComponentDesc sub : subComponents) {
            generateBindings(sub,pw);
        }
    }

    
    // ------------------------------------------------------------------
    // Implementation specific
    // ------------------------------------------------------------------

    /**
     * Return the type of the attribute set by the specified setter method name
     * in the specified interface.
     * 
     * @param itfName     the interface name
     * @param setterName  the setter method name
     * @return            the type of the attribute
     * @since 2.2.4
     */
    private UnifiedClass getAttrType( String itfName, String setterName )
    throws JuliacRuntimeException {

    	UnifiedClass uc = jc.create(itfName);
        UnifiedMethod[] ums = uc.getMethods();
        UnifiedMethod umeth = null;
        for (UnifiedMethod um : ums) {
			String umname = um.getName();
			if( umname.equals(setterName) ) {
				umeth = um;
				break;
			}
		}
        if( umeth == null ) {
        	String msg = "No setter method "+setterName+"in "+itfName;
			throw new JuliacRuntimeException(msg);
        }
        
		UnifiedClass[] ptypes = umeth.getParameterTypes();
		if( ptypes.length == 0 ) {
			String msg =
				"Method "+setterName+" in "+itfName+
				" is expected to declare at least one parameter";
			throw new JuliacRuntimeException(msg);
		}
    	
		return ptypes[0];
    }
}
