/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.desc;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;

/**
 * A struct-like class for holding data about a control membrane.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class MembraneDesc<T> {
    
    private String name;
    private String descriptor;
    private InterfaceType[] ctrlItfTypes;
    protected T ctrlDescs;
    protected ProxyClassGeneratorItf interceptorClassGenerator;
    
    /**
     * Create a new membrane descriptor.
     * 
     * @param name          the membrane name (only relevant for component-based membranes)
     * @param descriptor    the Julia membrane descriptor (e.g. primitive)
     * @param ctrlItfTypes  the array of control interface types implemented by this membrane
     * @param ctrlDescs     the structure of the membrane
     * @param jc            the instance of Juliac
     */
    public MembraneDesc(
        String name, String descriptor, InterfaceType[] ctrlItfTypes,
        T ctrlDescs ) {
        
        this.name = name;
        this.descriptor = descriptor;
        this.ctrlItfTypes = ctrlItfTypes;
        this.ctrlDescs = ctrlDescs;
    }
    
    /**
     * Return the membrane name (only relevant for component-based membranes).
     */
    public String getName() {
        return name;
    }
    
    /**
     * Return the Julia membrane descriptor (e.g. primitive).
     */
    public String getDescriptor() {
        return descriptor;
    }
    
    /**
     * Return the array of control interface types implemented by this membrane.
     */
    public InterfaceType[] getCtrlItfTypes() {
        return ctrlItfTypes;
    }
    
    /**
     * Return the list of controller descriptors associated with this membrane.
     */
    public T getCtrlDescs() {
        return ctrlDescs;
    }
    
    /**
     * Set the class generator for the interceptors associated with this
     * membrane.
     */
    public void setInterceptorClassGenerator( ProxyClassGeneratorItf icg ) {
        interceptorClassGenerator = icg;
    }

    /**
     * Return the class generator for the interceptors associated with this
     * membrane.
     */
    public ProxyClassGeneratorItf getInterceptorClassGenerator() {
        return interceptorClassGenerator;
    }
    
    /**
     * Return the implementation class of the controller implementing the
     * specified signature.
     * 
     * @param signature  the signature
     * @return           the class of the controller
     * @throws IllegalArgumentException  if there is no such controller
     * @since 2.2
     */
    public abstract UnifiedClass getCtrlImpl( String signature )
    throws IllegalArgumentException;
    
    @Override
    public String toString() {
        return descriptor;
    }
}
