/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.desc;

import java.util.List;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;

/**
 * A membrane descriptor where the structure of the membrane is represented by a
 * list of controller descriptors. This is the regular case for Julia-like
 * object-oriented control membranes.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2
 */
public class SimpleMembraneDesc extends MembraneDesc<List<ControllerDesc>> {

	private Juliac jc;
	
    /**
     * Create a new membrane descriptor.
     * 
     * @param jc            the instance of Juliac this descriptor is attached to
     * @param name          the membrane name (only relevant for component-based membranes)
     * @param descriptor    the Julia membrane descriptor (e.g. primitive)
     * @param ctrlItfTypes  the array of control interface types implemented by this membrane
     * @param ctrlDescs     the structure of the membrane
     */
    public SimpleMembraneDesc(
        Juliac jc, String name, String descriptor, InterfaceType[] ctrlItfTypes,
        List<ControllerDesc> ctrlDescs ) {
        
        super(name,descriptor,ctrlItfTypes,ctrlDescs);
        this.jc = jc;
    }
    /**
     * Return the implementation class of the controller implementing the
     * specified signature.
     * 
     * @param signature  the signature
     * @return           the class of the controller
     * @throws IllegalArgumentException  if there is no such controller
     * @since 2.2
     */
    public UnifiedClass getCtrlImpl( String signature )
    throws IllegalArgumentException {
        
        UnifiedClass ctrlItf = jc.create(signature);
        List<ControllerDesc> ctrlDescs = getCtrlDescs();
        
        for (ControllerDesc ctrlDesc : ctrlDescs) {
            String ctrlImplName = ctrlDesc.getImplName();
            UnifiedClass uc = jc.create(ctrlImplName);
            if( ctrlItf.isAssignableFrom(uc) ) {
                return uc;
            }
        }

        String msg =
            "No controller implementing "+signature+
            " in controller descriptor "+getDescriptor();
        throw new IllegalArgumentException(msg);
    }    
}
