/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.opt;

import static org.objectweb.fractal.juliac.visit.CodeHelper._new;
import static org.objectweb.fractal.juliac.visit.CodeHelper._throw;

import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Modifier;
import java.util.Date;
import java.util.List;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.Factory;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.conf.Constants;
import org.objectweb.fractal.juliac.desc.CompDesc;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.BoundedIOWriterMethodSourceCodeWriter;
import org.objectweb.fractal.juliac.visit.CatchSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;
import org.objectweb.fractal.util.Fractal;

/**
 * This class generates the source code of a component {@link
 * org.objectweb.fractal.api.factory.Factory}. This class is extended in
 * sub-packages with two concrete extensions: one which relies on a Fractal ADL
 * backend to parse an ADL file and one which visits a model of an ADL file.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 * @since 2.1.6
 */
public abstract class AbstractComponentFactoryClassGenerator
implements ClassGeneratorItf {
    
    private String targetname;
    
    public AbstractComponentFactoryClassGenerator( String targetname ) {        
        this.targetname = targetname;
    }
    
    /**
     * Return the name of the class generated by this generator.
     */
    public String getTargetClassName() {
        return targetname;
    }
    
    /**
     * Generate the header of the source code file associated with this
     * generator. The header is the section which precedes the source code of
     * the class. The header may be composed of an initial comment, the
     * declaration of the package name and some import clauses.
     * 
     * @param fv  the visitor for generating the header of the source code file
     */
    public void generate( FileSourceCodeVisitor fv ) throws IOException {
        
        int lastdot = targetname.lastIndexOf('.');
        String packagename = (lastdot==-1) ? null : targetname.substring(0,lastdot);
        
        /*
         * File header.
         */
        fv.visitFileHeader("/*");
        fv.visitFileHeader(" * Generated by "+getClass().getName()+" on: "+new Date());
        fv.visitFileHeader(" */");
        
        /*
         * Package.
         */
        if( packagename != null ) {
            fv.visitPackageName(packagename);
        }

        /*
         * Imports.
         */
        fv.visitImport(Component.class.getName());
        fv.visitImport(Fractal.class.getName());
        
        /*
         * Class.
         */
        ClassSourceCodeVisitor cv = fv.visitPublicClass();
        generate(cv);
    }
    
    /**
     * Generate the source code of the class associated with this generator.
     * 
     * @param cv  the visitor for generating the source code of the class
     */
    public void generate( ClassSourceCodeVisitor cv ) throws IOException {

        int lastdot = targetname.lastIndexOf('.');
        String classname = targetname.substring(lastdot+1);
        
        /*
         * Class declaration.
         */
        cv.visit(
            Modifier.PUBLIC, classname, null, null,
            new String[]{Factory.class.getName()} );
        
        /*
         * Empty implementation of the getFcContentDesc() method.
         */
        BlockSourceCodeVisitor mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, "Object", "getFcContentDesc", null,
                null );
        mv.visitIns(_throw(UnsupportedOperationException.class));
        mv.visitEnd();

        /*
         * Empty implementation of the getFcControllerDesc() method.
         */
        mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, "Object", "getFcControllerDesc", null,
                null );
        mv.visitIns(_throw(UnsupportedOperationException.class));
        mv.visitEnd();
        
        /*
         * Generate the newFcInstance999 methods.
         */
        BoundedIOWriterMethodSourceCodeWriter writer =
            new BoundedIOWriterMethodSourceCodeWriter(cv);
        PrintWriter pw = new PrintWriter(writer);
        CompDesc cdesc = generateNewFcInstanceMethods(pw,false);
        pw.close();
        writer.close();
        
        String root = cdesc.getID();
        ComponentType ct = cdesc.getCT();
        int newFcInstanceMethodCount = writer.getNewFcInstanceMethodCount();
        
        /*
         * Implementation of the newFcInstance() method.
         */
        mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, "Component", "newFcInstance", null,
                new String[]{InstantiationException.class.getName()} );
        CatchSourceCodeVisitor bloc = mv.visitTry();
        for (int i = 0; i < newFcInstanceMethodCount; i++) {
            bloc.visitIns("newFcInstance"+Integer.toString(i)+"()");
        }
        bloc = bloc.visitCatch(InstantiationException.class, "e");
        bloc.visitIns("throw","e");
        bloc = bloc.visitCatch(RuntimeException.class, "e");
        bloc.visitIns("throw","(RuntimeException)","e");
        bloc = bloc.visitCatch(Exception.class, "e");
        bloc.visitIns(_throw(ChainedInstantiationException.class, "e","null","e.getClass().getName()+\": \"+e.getMessage()"));
        bloc.visitEnd();
        mv.visitIns("return",root);
        mv.visitEnd();
        
        /*
         * One C999 field per component created by the newFcInstance() method.
         */
        List<String> compIDs = cdesc.getCompIDs();
        if( compIDs == null ) {
            String msg =
                "The compIDs field from CompDesc "+cdesc+
                " should have been set.";
            throw new NullPointerException(msg);
        }
        for (String compID : compIDs) {
            cv.visitField(
                Modifier.PRIVATE, Component.class.getName(), compID, null );
        }

        /*
         * Implementation of the getFcInstanceType() factory method which
         * returns the type of the component created by newFcInstance().
         */
        mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, Type.class.getName(),
                "getFcInstanceType", null, null );
        CatchSourceCodeVisitor block = mv.visitTry();
        block.visitComment("--------------------------------------------------");
        block.visitVar(ComponentType.class, "ct", Utils.javaify(ct).toString());
        block.visitComment("--------------------------------------------------");
        block.visitIns("return","ct");
        block = block.visitCatch(InstantiationException.class, "e");
        block.visitIns(_throw(Constants.JULIAC_RUNTIME_EXCEPTION, "e"));
        block.visitEnd();
        mv.visitEnd();
        
        /*
         * Implementation of the main(String[]) method which assumes that the
         * assembly provides a r/java.lang.Runnable method for running the
         * application. This method is provided as a convenience to avoid having
         * to use the o.o.f.juliac.runtime.Launcher class from the
         * juliac:runtime-oo module.
         */
        try {
            InterfaceType it = ct.getFcInterfaceType("r");
            String signature = it.getFcItfSignature();
            if( signature.equals(Runnable.class.getName()) ) {
                mv =
                    cv.visitMethod(
                        Modifier.PUBLIC + Modifier.STATIC, null, "void", "main",
                        new String[]{"String[] args"},
                        new String[]{
                            InstantiationException.class.getName(),
                            IllegalLifeCycleException.class.getName(),
                            NoSuchInterfaceException.class.getName()
                        } );
                
                // Instantiate the factory
                mv.visitVar(Factory.class,"f",_new(classname));

                // Instantiate the component
                mv.visitVar(Component.class,"c","f.newFcInstance()");

                // Start the component
                mv.visitIns(Fractal.class.getName()+".getLifeCycleController(c).startFc()");
                
                // Run the component
                mv.visitVar("Runnable","r","(Runnable)","c.getFcInterface(\"r\")");
                mv.visitIns("r.run()");

                // Stop the component
                mv.visitIns(Fractal.class.getName()+".getLifeCycleController(c).stopFc()");
                
                mv.visitEnd();
            }
        }
        catch (NoSuchInterfaceException e) {
            // No r method. Do not generate the main method.
        }
        
        cv.visitEnd();        
    }

    /**
     * Generate the <code>newFcInstance999</code> methods.
     * 
     * @param pw     the print writer where the factory is to be outputted
     * @param declareLocalVars
     *      <code>true</code> if the variables which hold references to
     *      components need to be declared
     */
    abstract protected CompDesc 
        generateNewFcInstanceMethods( PrintWriter pw, boolean declareLocalVars )
    throws IOException;
}
