/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt;

import java.lang.reflect.Modifier;
import java.util.Date;

import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;

/**
 * Default implementation for {@link ClassGeneratorItf}s.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.2
 */
public abstract class ClassGenerator implements ClassGeneratorItf {

    /**
     * Generate the header of the source code file associated with this
     * generator. The header is the section which precedes the source code of
     * the class. The header may be composed of an initial comment, the
     * declaration of the package name and some import clauses.
     * 
     * @param fv  the visitor for generating the header of the source code file
     */
    public void generate( FileSourceCodeVisitor fv ) {
        
        generateFileHeaders(fv);
        String packageName = getPackageName();
        if( packageName != null ) {
            fv.visitPackageName(packageName);
        }        
        generateImports(fv);
        
        ClassSourceCodeVisitor cv = fv.visitPublicClass();
        generate(cv);
    }
    
    /**
     * Generate the source code of the class associated with this generator.
     * 
     * @param cv  the visitor for generating the source code of the class
     */
    public void generate( ClassSourceCodeVisitor cv ) {
        
        int modifiers = getClassModifiers();
        String name = getClassName();
        String[] genericTypeParameters = getClassGenericTypeParameters();
        String superClassName = getSuperClassName();
        String[] implementedInterfaceNames = getImplementedInterfaceNames();
        cv.visit(
            modifiers, name, genericTypeParameters, superClassName,
            implementedInterfaceNames );

        generateFields(cv);
        generateConstructors(cv);
        generateMethods(cv);
        
        cv.visitEnd();
    }

    public void generateFileHeaders( FileSourceCodeVisitor fv ) {        
        fv.visitFileHeader("/*");
        fv.visitFileHeader(" * Generated by: "+getClass().getName());
        fv.visitFileHeader(" * on: "+new Date().toString());
        fv.visitFileHeader(" */");
    }

    public String getPackageName() {
        String targetclassname = getTargetClassName();
        int lastdot = targetclassname.lastIndexOf('.');
        if( lastdot == -1 ) {
            return null;
        }
        String packagename = targetclassname.substring(0,lastdot);
        return packagename;
    }
    
    public void generateImports( FileSourceCodeVisitor fv ) {}
    public int getClassModifiers() {
        return Modifier.PUBLIC;
    }

    public String getClassName() {
        String targetclassname = getTargetClassName();
        int lastdot = targetclassname.lastIndexOf('.');
        String shortclassname = targetclassname.substring(lastdot+1);
        return shortclassname;
    }

    public String[] getClassGenericTypeParameters() { return null; }
    public String getSuperClassName() { return null; }
    public String[] getImplementedInterfaceNames() { return null; }

    public void generateFields( ClassSourceCodeVisitor cv ) {}
    public void generateConstructors( ClassSourceCodeVisitor cv ) {}    
    public void generateMethods( ClassSourceCodeVisitor cv ) {}
}
