/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt;

import java.io.IOException;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.control.attribute.CloneableAttributeController;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.CfgLoaderItf;
import org.objectweb.fractal.juliac.api.MembraneLoaderItf;
import org.objectweb.fractal.juliac.conf.GeneratorMap;
import org.objectweb.fractal.juliac.conf.InterceptorSourceCodeGeneratorMap;
import org.objectweb.fractal.juliac.conf.CfgLoader;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.conf.JuliaLoader;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.AttributeControllerClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorClassGenerator;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;

/**
 * Default root class for source code generators associated to a given
 * optimization level.
 * 
 * @param <T>  the type of the structure representing the membrane
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.2
 */
public abstract class FCSourceCodeGenerator<T>
implements FCSourceCodeGeneratorItf<T> {
    
    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	protected Juliac jc;
    protected CfgLoaderItf jloader;
    protected MembraneLoaderItf mloader;
    
    public void init( Juliac jc ) throws IOException {

    	this.jc = jc;

    	// Post initialization phase
    	postInit();
    	
    	// Register the services
    	JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(FCSourceCodeGeneratorItf.class,this);
    }
    
    /**
     * This post initialization phase is kept separated from the {@link
     * #init()} method in order to be able to override in subclasses the
     * statements defined in this method.
     * 
     * @since 2.4
     */
    protected void postInit() throws IOException {
    	
    	// Initialize the membrane loader
    	JuliacConfig jconf = jc.getJuliacConfig();
		jloader =
			jconf.loadModuleIfNew(
				CfgLoader.class.getName(), CfgLoaderItf.class );
    	mloader = jconf.loadModule( JuliaLoader.class.getName() );
    	
    	// Post initialization phase for the JuliaLoader module
    	postInitJuliaLoaderModule();    	
    }
    
    /**
     * This post initialization phase is kept separated from the {@link
     * #init(Juliac)} method in order to be able to customize in subclasses the
     * <code>julia.config system</code> property used by this method.
     * 
     * @since 2.5
     */
    protected void postInitJuliaLoaderModule() throws IOException {

        /*
         * Load the configuration files referenced in the julia.config system
         * property.
         */
        String prop = System.getProperty("julia.config");
        if( prop!=null && prop.length()!=0 ) {
            String[] filenames = prop.split(",");
            for (String filename : filenames) {
                jloader.loadConfigFile(filename);
            }
        }
    }

    public void close( Juliac jc ) throws IOException {
    	// IOException thrown in subclasses (e.g. osgi/opt/oo)
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(FCSourceCodeGeneratorItf.class,this);
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified controller descriptor.
     * 
     * @param controllerDesc  the controller descriptor to be checked
     * @return                if <code>controllerDesc</code> should be accepted 
     * @since 2.2
     */
    public boolean acceptCtrlDesc( Object controllerDesc ) {
        
        if( ! (controllerDesc instanceof String) ) {
            return false;
        }
        String ctrlDesc = (String) controllerDesc;
        
        /*
         * Handle the prefix which may be specified in front of the controller
         * descriptor.
         */
        String prefix = Utils.getControllerDescPrefix(ctrlDesc);
        if( prefix != null ) {
        	if( ! this.prefix.equals(prefix) ) {
        		return false;
        	}
        	ctrlDesc = Utils.stripControllerDescPrefix(ctrlDesc);
        }
        
        /*
         * Check that the specified controller descriptor is defined in the .cfg
         * configuration files associated with Juliac.
         */
        boolean accept = mloader.containsKey(ctrlDesc);
        return accept;
    }
    
    /**
     * Set the prefix for controller descriptors associated with this generator.
     * This prefix is of the form <code>/prefix/</code> and may be specified
     * when several different optimization level source code generators are used
     * jointly to distinguish which optimization level should be used for a
     * particuler component, e.g. <code>/koch/primitive</code> to specify that
     * this is a primitive component that must be handled by the optimization
     * level source code generator associated with the <code>/koch/</code>
     * prefix.
     * 
     * @param prefix  the controller descriptor prefix
     * @since 2.5
     */
    public void setCtrlDescPrefix( String prefix ) {
    	this.prefix = prefix;
    }
    private String prefix = "";  // To ensure that prefix is never null
    
    /**
     * Generate the source code associated to a Fractal component.
     * The membrane implementation and the initializer implementation are
     * generated.
     * 
     * @param type            the component type
     * @param controllerDesc  the component controller descriptor
     * @param contentDesc     the component content descriptor
     * @return  the initializer class generator for this component
     */
    public InitializerClassGenerator<T> generate(
        Type type, Object controllerDesc, Object contentDesc )
    throws IOException {
        
        /*
         * Check parameter consistence. It's safe to perform the casts after
         * having performed the checks.
         */
        if( !(type instanceof ComponentType) ) {
            String msg = "ComponentType instance expected";
            throw new IllegalArgumentException(msg);
        }
        if( !(controllerDesc instanceof String) ) {
            String msg = "controllerDesc should be a String";
            throw new IllegalArgumentException(msg);
        }
        ComponentType ct = (ComponentType) type;
        String ctrlDesc = (String) controllerDesc;
                
        /*
         * Generate the membrane implementation.
         */
        String contentClassName =
        	Utils.getContentClassName(ctrlDesc,contentDesc);
        MembraneDesc<T> membraneDesc =
            generateMembraneImpl(ct,ctrlDesc,contentClassName);
        
        /*
         * Generate the initializer class.
         */
        InitializerClassGenerator<T> scg =
            generateInitializerImpl(membraneDesc,ct,contentDesc);
        return scg;
    }

    /**
     * Generate the source code of the interfaces and interceptors (hence
     * proxies) associated to the specified interface signature. Since
     * interceptors may depend of controller implementations (e.g. lifecycle
     * interceptors use lifecycle controllers), controllers need also to be
     * generated.
     * 
     * @param signature       the interface signature
     * @param controllerDesc  the component controller descriptor
     * @since 2.3.1
     */
    public void generateProxiesImpl( String signature, Object controllerDesc )
    throws IOException {
        
        /*
         * Check parameter consistence. It's safe to perform the casts after
         * having performed the checks.
         */
        if( !(controllerDesc instanceof String) ) {
            String msg = "controllerDesc should be a String";
            throw new IllegalArgumentException(msg);
        }
        String ctrlDesc = (String) controllerDesc;

        InterfaceType[] its =
        	new InterfaceType[]{
        		new BasicInterfaceType("none",signature,false,false,false)
        	};
        ComponentType ct = null;
        try {
			ct = new BasicComponentType(its);
		}
        catch (InstantiationException e) {
        	IOException ioe = new IOException();
        	ioe.initCause(e);
        	throw ioe;
		}
                
        /*
         * Generate the membrane implementation.
         */
        generateMembraneImpl(ct,ctrlDesc,null);
    }

    /**
     * Return the class generator for component interfaces.
     */
    public ProxyClassGeneratorItf getInterfaceClassGenerator( 
        InterfaceType it, MembraneDesc<?> membraneDesc ) {
        
        return new InterfaceImplementationClassGenerator(jc,it,membraneDesc,false);
    }
    

    // ------------------------------------------------------------------
    // Interceptor class and source code generator methods
    // ------------------------------------------------------------------
    
    /**
     * Return the interceptor class generator for the specified controller
     * descriptor and component type. Return <code>null</code> if no such
     * interceptor class generator exists. 
     */
    protected ProxyClassGeneratorItf getInterceptorClassGenerator(
        ComponentType ct, String ctrldesc ) {
        
        // Get the interceptor class generators
        boolean mergeable = false;
        InterceptorSourceCodeGeneratorItf[] scgs =
            getInterceptorSourceCodeGenerators(ct,ctrldesc,mergeable);
        if( scgs == null ) {
            // No interceptor
            return null;
        }
        
        // Prepare an InterceptorClassGenerator
        ProxyClassGeneratorItf icg =
        	getInterceptorClassGenerator(ct,ctrldesc,scgs);
        return icg;
    }
    
    /**
     * Return the class generator for interceptors.
     */
    protected ProxyClassGeneratorItf getInterceptorClassGenerator(
		ComponentType ct, String ctrldesc,
		InterceptorSourceCodeGeneratorItf[] iscgs ) {
    	
        String asm = mloader.getInterceptorClassGeneratorName(ct,ctrldesc);
        JuliacConfig jconf = jc.getJuliacConfig();
        GeneratorMap<InterceptorClassGenerator> iclgs =
    		jconf.interceptorClassGenerators();
        InterceptorClassGenerator icg = iclgs.get(asm);

    	// Initialize and return the interceptor class generator
    	icg.init(iscgs);
    	
    	return icg;
    }
    
    /**
     * Return the interceptor source code generators associated with the
     * specified component type and controller descriptor. Return
     * <code>null</code> if there is no such source code generator.
     */
    public InterceptorSourceCodeGeneratorItf[]
         getInterceptorSourceCodeGenerators(
    		 ComponentType ct, String ctrldesc, boolean mergeable ) {
    	
        String[] asms =
        	mloader.getInterceptorSourceCodeGeneratorNames(ct,ctrldesc);
        if( asms == null ) {
            // No interceptor
            return null;
        }
        
        JuliacConfig jconf = jc.getJuliacConfig();
        InterceptorSourceCodeGeneratorMap iscgs =
    		jconf.interceptorSourceCodeGenerators();

        return iscgs.get(asms,mergeable);
    }
    
    
    // ------------------------------------------------------------------
    // Implementation specific
    // ------------------------------------------------------------------
    
    /**
     * Return the class generator for component initializers. An initializer is
     * a factory and provides a <code>newFcInstance</code> method for
     * instantiating components
     */
    abstract protected InitializerClassGenerator<T>
        getInitializerClassGenerator();

    /**
     * Generate the implementation of the initializer class for the specified
     * component type.
     * 
     * @param membraneDesc       data about the membrane implementation
     * @param ct                 the component type
     * @param contentDesc        the component content descriptor
     */
    protected InitializerClassGenerator<T> generateInitializerImpl(
        MembraneDesc<T> membraneDesc, ComponentType ct, Object contentDesc )
    throws IOException {
        
        InitializerClassGenerator<T> scg = getInitializerClassGenerator();
        scg.init( jc, this, membraneDesc, ct, contentDesc );
        jc.generateSourceCode(scg);
        return scg;
    }
    
    /**
     * Generate the source code of component interfaces associated with the
     * specified array of interface types.
     */
    protected void generateInterfaceImpl( InterfaceType[] its, String ctrlDesc )
    throws IOException {
        for (InterfaceType it : its) {
            generateInterfaceImpl(it,ctrlDesc);
            InterfaceType internalit = Utils.newSymetricInterfaceType(it);
            generateInterfaceImpl(internalit,ctrlDesc);
        }
    }

    /**
     * Generate the source code of component interfaces associated with the
     * specified interface type.
     */
    protected void generateInterfaceImpl( InterfaceType it, String ctrlDesc )
    throws IOException {
        
        String signature = it.getFcItfSignature();
        UnifiedClass cl = jc.create(signature);
        UnifiedClass ac = jc.create(AttributeController.class.getName());
        UnifiedClass cac =
        	jc.create(CloneableAttributeController.class.getName());
        
        /*
         * If the interface is a business attribute control interface, generate
         * an implementation which provides a cloneFcAttributes method.
         * 
         * The second expression in the && skips cloneable attribute control
         * interfaces which are provided by parametric membranes.
         */
        if( ac.isAssignableFrom(cl) && ! cac.isAssignableFrom(cl) ) {
            ClassGeneratorItf ascg =
                new AttributeControllerClassGenerator(jc,it,null,false);
            jc.generateSourceCode(ascg);
        }
        
        ClassGeneratorItf iscg = getInterfaceClassGenerator(it,null);
        jc.generateSourceCode(iscg);
    }

    /**
     * Generate the implementation of interceptors associated to the specified
     * component type and the specified membrane descriptor.
     */
    protected void generateInterceptorImpl(
        InterfaceType[] its, MembraneDesc<?> membraneDesc )
    throws IOException {

    	ProxyClassGeneratorItf iscg =
            membraneDesc.getInterceptorClassGenerator();
        if( iscg == null ) {
            // No interceptor
            return;
        }
        
        // Generate the interceptors
        iscg.setJuliac(jc);
        iscg.setMembraneDesc(membraneDesc);
        for (InterfaceType it : its) {
            generateInterceptorImpl(it,iscg);
        }
    }
    
    /**
     * Generate the implementation of the interceptor associated to the
     * specified interface type using the specified interceptor source code
     * generator.
     */
    protected void generateInterceptorImpl(
        InterfaceType it, ProxyClassGeneratorItf iscg )
    throws IOException {
    	iscg.setJuliac(jc);
        iscg.setInterfaceType(it);
        if( iscg.match() ) {
            jc.generateSourceCode(iscg);
        }
    }
}
