/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.proxy;

import java.io.IOException;
import java.lang.reflect.Modifier;
import java.util.Date;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedClassHelper;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;

import com.google.common.base.Joiner;

/**
 * Root class for {@link InterfaceType}-based proxy class generators.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 */
public abstract class AbstractProxyClassGenerator
implements ProxyClassGeneratorItf {
    
	protected Juliac jc;
    protected InterfaceType it;
    protected MembraneDesc<?> membraneDesc;
    protected boolean mergeable;
    
    public AbstractProxyClassGenerator() {}
    
    public AbstractProxyClassGenerator(
        Juliac jc,
		InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable ) {
        
    	setJuliac(jc);
    	setInterfaceType(it);
        setMembraneDesc(membraneDesc);
        setMergeable(mergeable);
    }
    
    public void setInterfaceType( InterfaceType it ) {
        
    	/*
         * Check that the specified interface type corresponds to a Java
         * interface (not a class.)
         * 
         * Generating proxies based on class types leads to inconsistencies.
         * Indeed, when retrieving public methods, the methods from
         * java.lang.Object are returned. First, generating proxies for these
         * methods may seem counterintuitive. Second, and most importantly, some
         * of these methods (e.g. getClass) are final: trying to override them
         * in the proxy class leads to compilation errors.
         * 
         * This problem was raised by Philippe, Romain and Daniel in a case
         * where FraSCAti was erroneously trying to generate code for a service
         * interface of type Void.
         */
    	String signature = it.getFcItfSignature();
    	UnifiedClass uc = jc.create(signature);
    	if( ! uc.isInterface() ) {
    		String msg =
    			"Signature of InterfaceType "+signature+
    			" must correspond to a Java interface (not a class)";
    		throw new IllegalArgumentException(msg);
    	}
                
    	this.it = it;
	}
    
    public void setJuliac( Juliac jc ) { this.jc = jc; }    
    public void setMembraneDesc( MembraneDesc<?> membraneDesc ) { this.membraneDesc = membraneDesc; }
    public void setMergeable( boolean mergeable ) { this.mergeable = mergeable; }

    
    /**
     * Return the name of the class generated by this generator.
     */
    public String getTargetClassName() {
        String javaItfSignature = it.getFcItfSignature();
        String name = Utils.getJuliacGeneratedStrongTypeName(javaItfSignature);
        String suffix = getClassNameSuffix();
        String targetClassName = name + suffix;
        String rootedTargetClassName = jc.getRootedClassName(targetClassName);
        return rootedTargetClassName;
    }    

    /**
     * Generate the header of the source code file for the proxy class.
     * 
     * @param fv  the visitor for generating the header of the source code file
     */
    public void generate( FileSourceCodeVisitor fv ) throws IOException {
        
        // Exception is thrown by subclasses (e.g. see extension/felix)

        generateFileHeaders(fv);        
        String packageName = getPackageName();
        if( packageName != null ) {
            fv.visitPackageName(packageName);
        }
        generateImports(fv);
        
        ClassSourceCodeVisitor cv = fv.visitPublicClass();
        generate(cv);
    }
    
    /**
     * Generate the source code of the proxy class.
     * 
     * @param cv  the visitor where the source code is generated
     */
    public void generate( ClassSourceCodeVisitor cv ) {
        
        int modifiers = getClassModifiers();
        if( mergeable ) {
            /*
             * Do not define the class as public.
             * public was working with Spoon 20061202.
             * 
             * Between 1202 and 1230, the way Spoon deals with CompilationUnit
             * changed. A consequence is that the CtFile4SnippetSupport class
             * must provide a getPath() method (see
             * CtFile4SnippetSupport#getPath()).
             * 
             * If the class were public, JDT would check that the corresponding
             * file would exist (which is of course not the case as we deal with
             * snippets.)
             */
            modifiers = Modifier.ABSTRACT;
        }

        String name = getClassName();
        String[] genericTypeParameters = getClassGenericTypeParameters();
        String superClassName = getSuperClassName();
        String[] implementedInterfaceNames = getImplementedInterfaceNames();
        cv.visit(
            modifiers, name, genericTypeParameters, superClassName,
            implementedInterfaceNames );

        UnifiedClass proxycl = jc.create(it.getFcItfSignature());
        generateFields(cv,proxycl);
        generateConstructors(cv,proxycl);
        generateMethods(cv,proxycl);
        
        UnifiedMethod[] proxymethods =
        	UnifiedClassHelper.getMethodsWithMostSpecializedReturnType(proxycl);
        	
        for(UnifiedMethod proxym : proxymethods){
        	generateProxyMethod(cv,proxycl,proxym);
        }
        
        /*
         * Generate static blocks after fields as the Java compiler complains
         * ("Cannot reference a field before it is defined") if the static block
         * uses a static field which has not yet been defined.
         */
        generateStaticParts(cv,proxycl);

        cv.visitEnd();
    }
    
    public void generateFileHeaders( FileSourceCodeVisitor fv ) {        
        fv.visitFileHeader("/*");
        fv.visitFileHeader(" * Generated by: "+getClass().getName());
        fv.visitFileHeader(" * on: "+new Date().toString());
        fv.visitFileHeader(" */");
    }

    public String getPackageName() {
        String targetclassname = getTargetClassName();
        int lastdot = targetclassname.lastIndexOf('.');
        if( lastdot == -1 ) {
            return null;
        }
        String packagename = targetclassname.substring(0,lastdot);
        return packagename;
    }

    public void generateImports( FileSourceCodeVisitor fv ) {}
    public int getClassModifiers() { return Modifier.PUBLIC; }

    public String getClassName() {
        String targetclassname = getTargetClassName();
        int lastdot = targetclassname.lastIndexOf('.');
        String shortclassname = targetclassname.substring(lastdot+1);
        return shortclassname;
    }

    public String[] getClassGenericTypeParameters() {
        UnifiedClass proxycl = jc.create(it.getFcItfSignature());
        String[] tps = proxycl.getTypeParameters();
        return tps;
    }
    
    public String getSuperClassName() { return null; }
    public String[] getImplementedInterfaceNames() { return null; }

    public void generateStaticParts( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {}
    public void generateFields( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {}
    public void generateConstructors( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {}
    public void generateMethods( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {}

    public void generateProxyMethod(
        ClassSourceCodeVisitor cv, UnifiedClass proxycl, UnifiedMethod proxym ) {

        int modifiers = Modifier.PUBLIC;
        String[] genericReturnType = proxym.getTypeParameters();
        String returnType = proxym.getGenericReturnType();
        String name = proxym.getName();
        String[] parameters = proxym.getParameters();
        String[] exceptions = proxym.getGenericExceptionTypes();
        
        /*
         * Declare all method parameters as final.
         * Needed when the parameter is used in an anonymous inner class.
         * Used by Tinfi for oneway methods (Thread created.)
         */
        for (int i = 0; i < parameters.length; i++) {
            if( ! parameters[i].startsWith("final") ) {
                parameters[i] = "final "+parameters[i];
            }
        }
        
        if(mergeable) {
            // _super_ method declaration to propagate the invocation to preceding layers
            int mergeablemodifiers = Modifier.PUBLIC + Modifier.ABSTRACT;
            String mergeablename = "_super_"+name;
            BlockSourceCodeVisitor mv =
                cv.visitMethod(
                    mergeablemodifiers, genericReturnType, returnType,
                    mergeablename, parameters, exceptions );
            mv.visitEnd();
        }

        // Method declaration
        BlockSourceCodeVisitor mv =
            cv.visitMethod(
                modifiers, genericReturnType, returnType,
                name, parameters, exceptions );
        
        // Method body
        generateProxyMethodBody(mv,proxycl,proxym);
        mv.visitEnd();
    }
    
    public void generateProxyMethodBody(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
                
        // Before code
        generateProxyMethodBodyBeforeCode(mv,proxycl,proxym);
        
        // Proceed with the call to the delegate
        generateProxyMethodBodyDelegatingCode(mv,proxycl,proxym);
        
        // After code
        generateProxyMethodBodyAfterCode(mv,proxycl,proxym);
        
        // Return the result
        String rtypename = proxym.getGenericReturnType();
        if( ! rtypename.equals("void") ) {
            mv.visitIns("return","ret");
        }
        
        // Finally code
        generateProxyMethodBodyFinallyCode(mv,proxycl,proxym);
    }
    
    public void generateProxyMethodBodyBeforeCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {}
    
    public void generateProxyMethodBodyDelegatingCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        UnifiedClass[] ptypes = proxym.getParameterTypes();
        String rtypename = proxym.getGenericReturnType();
        String[] tpnames = proxym.getTypeParameterNames();

        String line = "";
        if( ! rtypename.equals("void") ) {
           line += rtypename+" ret = ";
        }
        
        String delegatingInstance = getDelegatingInstance(proxycl, proxym);
        String delegatingMethodName = getDelegatingMethodName(proxycl,proxym);
        line+= delegatingInstance+".";
        if( tpnames.length > 0 ) {
        	line+="<"+Joiner.on(",").join(tpnames)+">";
        }
        line+=delegatingMethodName+"(";
        for (int j = 0; j < ptypes.length; j++) {
            if(j!=0)  line+=",";
            line+="arg"+Integer.toString(j);
        }
        mv.visitIns(line+")");
    }
    
    public void generateProxyMethodBodyAfterCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {}
    public void generateProxyMethodBodyFinallyCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {}

    public String getDelegatingMethodName( UnifiedClass proxycl, UnifiedMethod proxym ) {
        String name = proxym.getName();
        if(mergeable) {
            return "_super_" + name;
        }
        return name;
    }
}
