/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.proxy;

import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.FileSourceCodeVisitor;

/**
 * This class is a {@link ProxyClassGenerator} which aggregates the
 * source code generated by several {@link InterceptorSourceCodeGeneratorItf}s.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 */
public class InterceptorClassGenerator extends AbstractProxyClassGenerator {
    
    /**
     * The {@link InterceptorSourceCodeGeneratorItf}s managed by this
     * generator.
     */
	protected InterceptorSourceCodeGeneratorItf[] iscgs;

    /**
     * Each element in this array is <code>true</code> if the corresponding
     * {@link InterceptorSourceCodeGeneratorItf} managed by this generator
     * applies in the current state.
     */
    protected boolean[] matches;
    
    public InterceptorClassGenerator() {
    	super();
    }

    public InterceptorClassGenerator(
        Juliac jc, InterceptorSourceCodeGeneratorItf[] iscgs,
        InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable ) {
        
        init(iscgs);
        
        /*
         * Don't call super(jc,it,membraneDesc,mergeable)
         * Else, the setting of it, membraneDesc, mergeable and jc won't be
         * propagated to all elements of iscgs.
         * 
         * jc must be set first as this is the case in the inherited
         * constructor. The reason is that the inherited setInterfaceType method
         * uses jc.
         */
        setJuliac(jc);
        setInterfaceType(it);
        setMembraneDesc(membraneDesc);
        setMergeable(mergeable);
    }
    
    public void init( InterceptorSourceCodeGeneratorItf[] iscgs )
    throws IllegalArgumentException {
        
        if( iscgs == null ) {
            String msg = "Parameter should not be null";
            throw new IllegalArgumentException(msg);
        }
        
        this.iscgs = iscgs;
        this.matches = new boolean[this.iscgs.length];
    }
    
    // --------------------------------------------------------------------
    // State management
    // --------------------------------------------------------------------
    
    @Override
    public void setInterfaceType( InterfaceType it ) {
        super.setInterfaceType(it);
        if( iscgs != null ) {
            for (InterceptorSourceCodeGeneratorItf iscg : iscgs) {
                iscg.setInterfaceType(it);
            }
            updateState();
        }
    }
    
    @Override
    public void setMembraneDesc( MembraneDesc<?> membraneDesc ) {
        super.setMembraneDesc(membraneDesc);
        if( iscgs != null ) {
            for (InterceptorSourceCodeGeneratorItf iscg : iscgs) {
                iscg.setMembraneDesc(membraneDesc);
            }
        }
    }
    
    @Override
    public void setMergeable( boolean mergeable ) {
        super.setMergeable(mergeable);
        if( iscgs != null ) {
            for (InterceptorSourceCodeGeneratorItf iscg : iscgs) {
                iscg.setMergeable(mergeable);
            }
        }
    }
    
    private void updateState() {
        for (int i = 0; i < iscgs.length; i++) {
            matches[i] = iscgs[i].match();
        }
    }

    // --------------------------------------------------------------------
    // ProxyClassGeneratorItf methods
    // --------------------------------------------------------------------
    
    @Override
    public void generateImports( FileSourceCodeVisitor fv ) {
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateImports(fv);
            }
        }
    }
    
    @Override
    public void generateConstructors( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        if( ! mergeable ) {
        	
            // public constructor()
            BlockSourceCodeVisitor mv = cv.visitConstructor(Modifier.PUBLIC,null,null,null);
            mv.visitEnd();

            // private constructor(Object) - used by the clone() method only
            mv = cv.visitConstructor(Modifier.PRIVATE,null,new String[]{"Object obj"},null);
            mv.visitIns("setFcItfDelegate(obj)");
            mv.visitEnd();
        }
    }

    @Override
    public void generateFields( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        
        generateFieldImpl(cv,proxycl);

        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateFields(cv, proxycl);
            }
        }
    }

    @Override
    public void generateMethods( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        generateMethodInitFcController(cv,proxycl);
        generateMethodClone(cv,proxycl);
        generateMethodGetFcItfDelegate(cv,proxycl);
        generateMethodSetFcItfDelegate(cv,proxycl);
        generateMethodOthers(cv,proxycl);
    }
    
    @Override
    public void generateProxyMethodBodyAfterCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym) {
        for (int i = iscgs.length-1; i >= 0; i--) {
            if( matches[i] ) {
                iscgs[i].generateProxyMethodBodyAfterCode(mv,proxycl,proxym);
            }
        }
    }

    @Override
    public void generateProxyMethodBodyBeforeCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym) {
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateProxyMethodBodyBeforeCode(mv,proxycl,proxym);
            }
        }
    }

    @Override
    public void generateStaticParts( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateStaticParts(cv,proxycl);
            }
        }
    }
    
    @Override
    public void generateProxyMethodBodyFinallyCode(
        BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym) {
        for (int i = iscgs.length-1; i >= 0; i--) {
            if( matches[i] ) {
                iscgs[i].generateProxyMethodBodyFinallyCode(mv,proxycl,proxym);
            }
        }
    }

    @Override
    public String[] getClassGenericTypeParameters() {
    	/*
    	 * We don't need to concatenate the generic type parameters of all
    	 * interceptor source code generators which match. Since they are all
    	 * associated with the same interface type, they return the same array.
    	 * This is then sufficient to return the array of the first one which
    	 * matches.
    	 */
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                String[] names = iscgs[i].getClassGenericTypeParameters();
                if( names != null ) {
                	return names;
                }
            }
        }
        return null;
    }

    @Override
    public int getClassModifiers() {
        int result = 0;
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                int mods = iscgs[i].getClassModifiers();
                result |= mods;
            }
        }
        return result;
    }

    public String getDelegatingInstance( UnifiedClass proxycl, UnifiedMethod proxym ) {
        return mergeable ? "this" : "_impl";
    }

    @Override
    public String[] getImplementedInterfaceNames() {
        
        Set<String> all = new HashSet<String>();
        boolean allnull = true;
        
        if( mergeable ) {
            /*
             * When declared as mergeable, the layer does not implement
             * Interceptor. It remains to be seen whether this choice should be
             * reconsidered in the future.
             */
        }
        else {
            all.add( Interceptor.class.getName() );
            allnull = false;
        }
        
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                String[] names = iscgs[i].getImplementedInterfaceNames();
                if( names != null ) {
                    allnull = false;
                    all.addAll(Arrays.asList(names));
                }
            }
        }
        if(allnull) { return null; }
        String[] result = all.toArray(new String[all.size()]);
        return result;
    }

    public String getClassNameSuffix() {
        StringBuffer sb = new StringBuffer();
        sb.append("Interceptor");
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                String name = iscgs[i].getClassNameSuffix();
                sb.append(name);
            }
        }
        return sb.toString();
    }

    @Override
    public String getSuperClassName() throws IllegalArgumentException {
        String superClassName = null;
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                String name = iscgs[i].getSuperClassName();
                if( name != null ) {
                    if( superClassName == null ) {
                        superClassName = name;
                    }
                    else {
                        if( ! superClassName.equals(name) ) {
                            String msg =
                                "Interceptor class can not extend both "+
                                superClassName+" and "+name;
                            throw new IllegalArgumentException(msg);
                        }
                    }
                }
            }
        }
        return superClassName;
    }
    
    public boolean match() {
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                return true;
            }
        }
        return false;
    }

    
    // --------------------------------------------------------------------
    // InterceptorClassGenerator methods
    // --------------------------------------------------------------------
    
    public void generateFieldImpl( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {        
        if( ! mergeable ) {
            String signature = jc.getInterfaceTypeSignature(it);
            String impl = getDelegatingInstance(proxycl,null);
            cv.visitField(Modifier.PRIVATE, signature, impl, null);
        }
    }

    public void generateMethodInitFcController( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        if( ! mergeable ) {
            // Method Controller.initFcController
            BlockSourceCodeVisitor mv =
                cv.visitMethod(
                    Modifier.PUBLIC, null, "void", "initFcController",
                    new String[]{InitializationContext.class.getName()+" ic"},
                    new String[]{InstantiationException.class.getName()} );
            generateMethodInitFcController(mv,proxycl);
            mv.visitEnd();
        }
    }
        
    public void generateMethodInitFcController( BlockSourceCodeVisitor mv, UnifiedClass proxycl ) {
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateMethodInitFcController(mv,proxycl);
            }
        }
    }
    
    public void generateMethodClone( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        if( ! mergeable ) {
            // Method Object.clone
            BlockSourceCodeVisitor mv =
                cv.visitMethod(Modifier.PUBLIC, null, "Object", "clone", null, null );
            generateMethodClone(mv,proxycl);
            mv.visitEnd();
        }
    }
    
    public void generateMethodClone( BlockSourceCodeVisitor mv, UnifiedClass proxycl ) {
        // TODO generate generic parameter names (e.g. IInterceptorLC<W,K,..> when appropriate)
        String className = getClassName();

        mv.visitVar(className,"clone","new",className+"(getFcItfDelegate())");
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateMethodClone(mv,proxycl);
            }
        }
        mv.visitIns("return","clone");    	
    }
    
    public void generateMethodGetFcItfDelegate(ClassSourceCodeVisitor cv, UnifiedClass proxycl) {        
        if( ! mergeable ) {
            // Method Interceptor.getFcItfDelegate
            BlockSourceCodeVisitor mv =
                cv.visitMethod(Modifier.PUBLIC, null, "Object", "getFcItfDelegate", null, null );
            generateMethodGetFcItfDelegate(mv,proxycl);
            mv.visitEnd();
        }
    }

    public void generateMethodGetFcItfDelegate(BlockSourceCodeVisitor mv, UnifiedClass proxycl) {        
        String impl = getDelegatingInstance(proxycl,null);
        mv.visitIns("return",impl);
    }
    
    public void generateMethodSetFcItfDelegate(ClassSourceCodeVisitor cv, UnifiedClass proxycl) {
        if( ! mergeable ) {
            // Method Interceptor.setFcItfDelegate
            BlockSourceCodeVisitor mv =
                cv.visitMethod(
            		Modifier.PUBLIC, null, "void", "setFcItfDelegate",
            		new String[]{"Object obj"}, null );
            generateMethodSetFcItfDelegate(mv,proxycl);
            mv.visitEnd();
        }
    }

    public void generateMethodSetFcItfDelegate(BlockSourceCodeVisitor mv, UnifiedClass proxycl) {
        String impl = getDelegatingInstance(proxycl,null);
        mv.visitSet(impl,"("+it.getFcItfSignature()+")obj");
    }
    
    public void generateMethodOthers(ClassSourceCodeVisitor cv, UnifiedClass proxycl) {        
        for (int i = 0; i < matches.length; i++) {
            if( matches[i] ) {
                iscgs[i].generateMethodOthers(cv,proxycl);
            }
        }
    }
}
