/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 */

package org.objectweb.fractal.juliac.ucf;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;

/**
 * Implementation of {@link UnifiedMethod} for {@link Method}s.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @since 1.3
 */
public class UCMethod implements UnifiedMethod {
    
    /** The concrete {@link Method}. */
    private Method method;
    
    /**
     * The class declaring the current method.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
    private UClass uClass;
    
    public UCMethod( UClass uClass, Method method ) {
    	this.uClass = uClass;
        this.method = method;
    }

    /**
     * Return the class declaring the current method.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
	public UClass getDeclaringClass() {
		return uClass;
	}
	
    /**
     * Return the method annotation for the specified annotation type,
     * or <code>null</code> if the method is not annotated.
     */
    public <A extends Annotation> A getAnnotation( Class<A> annotationType ) {
        A a = method.getAnnotation(annotationType);
        return a;
    }
    
    public Annotation[] getAnnotations() {
    	Annotation[] annots = method.getAnnotations();
    	return annots;
    }

    /**
     * Return the types of the exceptions thrown by the current method.
     */
    public UnifiedClass[] getExceptionTypes() {
        Class<?>[] params = method.getExceptionTypes();
        UnifiedClass[] ucs = new UnifiedClass[ params.length ];
        for (int i = 0; i < params.length; i++) {
            ucs[i] = new UClass(params[i]);
        }
        return ucs;
    }

    /**
     * Return the string representations of the generic exception types.
     * 
     * @since 2.1
     */
    public String[] getGenericExceptionTypes() {
        Type[] throwns = method.getGenericExceptionTypes();
        String[] gets = new String[ throwns.length ];
        for (int i = 0; i < throwns.length; i++) {
            String typename = this.uClass.toString(throwns[i]);
            gets[i] = typename;
        }
        return gets;
    }
    
    /**
     * Return a string representation of the generic return type.
     * 
     * @since 2.1
     */
    public String getGenericReturnType() {
        Type type = method.getGenericReturnType();
        String typename = this.uClass.toString(type);
        return typename;
    }
    
    /**
     * Return the modifiers of the current method.
     */
    public int getModifiers() {
        int result = method.getModifiers();
        return result;
    }
    
    /**
     * Return the name of the current method.
     */
    public String getName() {
        return method.getName();
    }

    /**
     * Return a string representation (the parameter type followed by the space
     * character, followed by the type name) for the current method.
     */
    public String[] getParameters() {
        Type[] tps = method.getGenericParameterTypes();
        String[] params = new String[ tps.length ];
        for (int i = 0; i < tps.length; i++) {
            String typename = uClass.toString(tps[i]);
            StringBuffer sb = new StringBuffer(typename);
            sb.append(" arg");
            sb.append(i);
            params[i] = sb.toString();
        }
        return params;
    }
    
    /**
     * Return the types of the parameters of the current method.
     */
    public UnifiedClass[] getParameterTypes() {
        Class<?>[] params = method.getParameterTypes();
        UnifiedClass[] ucs = new UnifiedClass[ params.length ];
        for (int i = 0; i < params.length; i++) {
            ucs[i] = new UClass(params[i]);
        }
        return ucs;
    }

    /**
     * Return the return type of the current method.
     */
    public UnifiedClass getReturnType() {
        Class<?> cl = method.getReturnType();
        UnifiedClass uc = new UClass(cl);
        return uc;
    }

    /**
     * Return a string representation of the current method based on the method
     * name and its parameter types such as <code>name(type1,..typen)</code>.
     */
    @Override
    public String toString() {
        
    	/*
    	 * Do not modify the implementation of this method unless you are sure
    	 * that this does not break
    	 * AbstractProxyClassGenerator#generate(ClassSourceCodeVisitor)
    	 * 
    	 * See the skipping which is implemented for subclasses which redefine
    	 * inherited methods with a more specialized return type.
    	 * 
    	 * This is checked by examples/advanced/foobar-cases/foobar-cases.
    	 */
        
    	StringBuffer sb = new StringBuffer();
    	
        sb.append(" ");
        sb.append(getName());
        
        String[] params = getParameters();
        sb.append('(');
        for (int i = 0; i < params.length; i++) {
        	String param = params[i];
            if(i!=0) { sb.append(','); }
            sb.append(param);
        }        
        sb.append(')');
        
        String[] exceptions = getGenericExceptionTypes();
        if(exceptions!=null && exceptions.length!=0){
        	sb.append(" throws ");
        	for(int i=0; i<exceptions.length; i++){
        		if(i!=0){
        			sb.append(", ");        			
        		}
        		sb.append(exceptions[i]);
        	}        	
        }
                
        sb.append(";");
        
        return sb.toString();
    }
    
    /**
     * Return the names of the type variables declared by the generic
     * declaration for the current method. Return an empty array if the current
     * method does not declare any type variable.
     * 
     * @since 2.2.6
     */
    public String[] getTypeParameterNames() {
    	
        TypeVariable<Method>[] tvs = method.getTypeParameters();
        if( tvs.length == 0 ) {
            return new String[0];
        }
        
        String[] tps = new String[ tvs.length ];
        for (int i = 0; i < tvs.length; i++) {
            TypeVariable<Method> tv = tvs[i];
            tps[i] = tv.getName();
        }
        
        return tps;
    }
    
    /**
     * Return the string representations of the type variables declared by the
     * generic declaration for the current method. Return an empty array if the
     * current method does not declare any type variable.
     * 
     * @since 2.1
     */
    public String[] getTypeParameters() {
        
        TypeVariable<Method>[] tvs = method.getTypeParameters();
        if( tvs.length == 0 ) {
            return new String[0];
        }
        
        String[] tps = new String[ tvs.length ];
        for (int i = 0; i < tvs.length; i++) {
            TypeVariable<Method> tv = tvs[i];
            
            Type[] bounds = tv.getBounds();
            if( bounds.length == 0 ) {
                // Given the specification of getBounds(), this shouldn't occur
                String msg = "Unsupported array bounds of length 0 for: "+tv;
                throw new RuntimeException(msg);
            }
            
            StringBuffer sb = new StringBuffer( tv.getName() );
            if( bounds.length!=1 || ! Object.class.equals(bounds[0]) ) {
                sb.append(" extends ");
                for (int j = 0; j < bounds.length; j++) {
                    if(j!=0) sb.append(" & ");
                    String s = uClass.toString(bounds[j]);
                    sb.append(s);
                }
            }
            tps[i] = sb.toString();
        }
        
        return tps;
    }
}
