/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 */

package org.objectweb.fractal.juliac.ucf;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.RuntimeClassNotFoundException;
import org.objectweb.fractal.juliac.conf.JuliacConfig;

/**
 * Implementation of {@link UnifiedClass} for {@link Class}es.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @since 1.3
 */
public class UClass implements UnifiedClass {
    
	private Class<?> cl;
    private Juliac jc;

    /**
     * Map from generic types to their instantiated values. 
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
    private Map<Type,Type> genericTypeValues = new HashMap<Type,Type>();    
    
    public UClass( String name, Juliac jc )
    throws RuntimeClassNotFoundException {

        this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        ClassLoader classloader = jconf.getClassLoader();
        
        try {
        	Class<?> cl = classloader.loadClass(name);
            setClass(cl);
        }
        catch( ClassNotFoundException e ) {
            if( primitives.containsKey(name) ) {
                cl = primitives.get(name);
            }
            else {
                throw new RuntimeClassNotFoundException(name);
            }
        }        
    }
    
    final private static Map<String,Class<?>> primitives =
        new HashMap<String,Class<?>>() {
            private static final long serialVersionUID = 4126982862148826431L;
        {
            put("boolean",boolean.class);
            put("char",char.class);
            put("byte",byte.class);
            put("short",short.class);
            put("int",int.class);
            put("long",long.class);
            put("float",float.class);
            put("double",double.class);
            put("void",void.class);
        }};
    
    final private static Map<Class<?>,UClass> boxed =
        new HashMap<Class<?>,UClass>() {
            private static final long serialVersionUID = 2398856843237335831L;
        {
            put(boolean.class,new UClass(Boolean.class));
            put(char.class,new UClass(Character.class));
            put(byte.class,new UClass(Byte.class));
            put(short.class,new UClass(Short.class));
            put(int.class,new UClass(Integer.class));
            put(long.class,new UClass(Long.class));
            put(float.class,new UClass(Float.class));
            put(double.class,new UClass(Double.class));
            put(void.class,new UClass(Void.class));
        }};
        
    final private static Map<Class<?>,String> nulvalues =
        new HashMap<Class<?>,String>() {
            private static final long serialVersionUID = -8428019352784119543L;
        {
            put(boolean.class,"false");
            put(char.class," ");
            put(byte.class,"0");
            put(short.class,"0");
            put(int.class,"0");
            put(long.class,"0");
            put(float.class,"0.0f");
            put(double.class,"0.0");
        }};
        
    public UClass( Class<?> cl ) {
        this.setClass(cl);
    }
    
    /**
     * Setter for the {@link #cl} field. This method updates the map of generic
     * type value {@link #genericTypeValues}.
     * 
     * @param cl  class declaration associated to this UnifiedClass
     * 
     * @since 2.2.5
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     */
	private void setClass( Class<?> cl ) {
    	
		this.cl = cl;
    	
		if( cl.isPrimitive() ) {
			// Primitive types can not declare generics
    		return;
    	}
    	
    	Stack<Class<?>> genericInterfaces = new Stack<Class<?>>();
		Set<Class<?>> interfacesAlreadyAnalized = new HashSet<Class<?>>();
		
		genericInterfaces.add(cl);
		
		while( ! genericInterfaces.isEmpty() ) {
			Class<?> genericInterface = genericInterfaces.pop();
			Type[] ginterfaces = genericInterface.getGenericInterfaces();
			for( Type ginterface : ginterfaces ) {
				if( ginterface instanceof ParameterizedType ){
					ParameterizedType parameterizedType = (ParameterizedType) ginterface;					
					Class<?> rawType = (Class<?>)parameterizedType.getRawType();
					if( ! interfacesAlreadyAnalized.add(rawType) ) {
						continue;
					}
					TypeVariable<?>[] rawTypeTypeParameters = rawType.getTypeParameters();
					Type[] typeArguments = parameterizedType.getActualTypeArguments();
					for(int i=0; i<rawTypeTypeParameters.length; i++){
						TypeVariable<?> rawTypeTypeParameter = rawTypeTypeParameters[i];
						Type typeArgument = typeArguments[i];
						Type typeVariableValue = genericTypeValues.get(typeArgument);
						if( typeVariableValue == null ){
							genericTypeValues.put(rawTypeTypeParameter, typeArgument);
						}
						else {
							genericTypeValues.put(rawTypeTypeParameter, typeVariableValue);
						}
					}
					genericInterfaces.push(rawType);					
				}
				else if( ginterface instanceof Class<?> ){
					genericInterfaces.push((Class<?>)ginterface);
				}
				else {
					String msg =
						"SuperGenericInterface is not a class : " +
						ginterface + "(" +
						ginterface.getClass() + ")";
					throw new JuliacRuntimeException(msg);
				}
			}
		}		
    }
    
    public UnifiedClass box() {
        if( cl.isPrimitive() ) {
            UClass b = boxed.get(cl);
            return b;
        }
        else {
            return this;
        }
    }

    public boolean isAssignableFrom( UnifiedClass uc ) {
        try {
            UClass other = toUClass(uc);
            boolean b = cl.isAssignableFrom(other.cl);
            return b;
        }
        catch( RuntimeClassNotFoundException ucnfe ) {
            /*
             * In this case, try to perform the reverse test. This gives a
             * chance to uc to try to transform the current object to an object
             * implementing its type.
             */
            boolean b = uc.isAssignableTo(this);
            return b;
        }
    }
    
    public boolean isAssignableTo( UnifiedClass uc ) {
        try {
            UClass other = toUClass(uc);
            boolean b = other.cl.isAssignableFrom(cl);
            return b;
        }
        catch( RuntimeClassNotFoundException ucnfe ) {
            /*
             * To not try to perform the reverse test as in
             * #isAssignableFrom(UnifiedClass) as this would lead to infinite
             * recursion.
             */
            return false;
        }
    }
    
    public boolean isInterface() {
        boolean isInterface = cl.isInterface();
        return isInterface;
    }

    public boolean isPrimitive() {
        boolean isPrimitive = cl.isPrimitive();
        return isPrimitive;
    }
    
    public <A extends Annotation> A getAnnotation( Class<A> annotationType ) {
        A a = cl.getAnnotation(annotationType);
        return a;
    }

    public Annotation[] getAnnotations() {
    	Annotation[] annots = cl.getAnnotations();
    	return annots;
    }
    
    public UnifiedClass getDeclaringClass() {
        Class<?> dc = cl.getDeclaringClass();
        if( dc == null ) {
            // No declaring class
            return null;
        }
        UnifiedClass udc = new UClass(dc);
        return udc;
    }

    public UnifiedField[] getDeclaredFields() {
        Field[] fields = cl.getDeclaredFields();
        UnifiedField[] ufields = new UnifiedField[ fields.length ];
        for (int i = 0; i < fields.length; i++) {
            ufields[i] = new UCField(this,fields[i]);
        }
        return ufields;
    }
    
    public UnifiedClass[] getInterfaces() {
    	Class<?>[] classes = cl.getInterfaces();
    	UnifiedClass[] ucs = new UnifiedClass[ classes.length ];
    	for (int i = 0; i < classes.length; i++) {
			ucs[i] = new UClass(classes[i]);
		}
    	return ucs;
    }
    
    public UnifiedField getDeclaredField( String name )
    throws NoSuchFieldException {
    	Field field = cl.getDeclaredField(name);
    	UnifiedField ufield = new UCField(this,field);
    	return ufield;
    }
    
    public UnifiedField[] getFields() {
        Field[] fields = cl.getFields();
        UnifiedField[] ufields = new UnifiedField[ fields.length ];
        for (int i = 0; i < fields.length; i++) {
            ufields[i] = new UCField(this,fields[i]);
        }
        return ufields;
    }
    
    public UnifiedMethod[] getMethods() {          
        Method[] meths = cl.getMethods();
        UnifiedMethod[] umeths = new UnifiedMethod[ meths.length ];
        for (int i = 0; i < meths.length; i++) {
            umeths[i] = new UCMethod(this,meths[i]);
        }
        return umeths;
    }
    
    public UnifiedMethod getMethod( String name, Class<?>... types )
    throws NoSuchMethodException {
        Method m = cl.getMethod(name,types);
        UnifiedMethod um = new UCMethod(this, m);
        return um;
    }
    
    public int getModifiers() {
        int mods = cl.getModifiers();
        return mods;
    }
    
    public String getName() {
        return getName(cl);
    }
    
    public String getNullValue() {
        if( nulvalues.containsKey(cl) ) {
            String ret = nulvalues.get(cl);
            return ret;
        }
        return "null";
    }
    
    public UnifiedClass getSuperclass() {
    	Class<?> sup = cl.getSuperclass();
    	UnifiedClass uc = new UClass(sup);
    	return uc;
    }
    
    /**
     * Return the string representations of the type variables declared by the
     * current type.
     * 
     * @since 2.1.1
     */
    public String[] getTypeParameters() {
        TypeVariable<?>[] tvs = cl.getTypeParameters();
        String[] rets = new String[tvs.length];
        for (int i = 0; i < tvs.length; i++) {
            StringBuffer sb = new StringBuffer();
            String s = tvs[i].toString();
            sb.append(s);
            Type[] bounds = tvs[i].getBounds();
            if( ! (bounds.length==0 || bounds[0].equals(Object.class) ) ) {
                sb.append(" extends ");
                for (int j = 0; j < bounds.length; j++) {
                    if(j!=0) sb.append(" & ");
                    Type bound = bounds[j];
                    s = toString(bound);
                    sb.append(s);
                }
            }
            rets[i] = sb.toString();
        }
        return rets;
    }

    /**
     * Return the string representations of the type variable names declared by
     * the current type. This method differs from {@link #getTypeParameters()}
     * by omitting upper bounds in the return type variables. Upper bounds are
     * elements such as <code>extends FooBar</code> in a generic declaration of
     * the form <code>&lt;T extends FooBar&gt;</code>.
     * 
     * @since 2.1.1
     */
    public String[] getTypeParameterNames() {
        TypeVariable<?>[] tvs = cl.getTypeParameters();
        String[] rets = new String[tvs.length];
        for (int i = 0; i < tvs.length; i++) {
            rets[i] = tvs[i].toString();
        }
        return rets;
    }
    
    public String getValueDelimiter() {
    	if( cl.equals(String.class) ) {
    		return "\"";
    	}
    	if( cl.equals(char.class) || cl.equals(Character.class) ) {
    		return "'";
    	}
    	return "";
    }
    
    /**
     * @since 2.2.6
     */
    @Override
    public boolean equals( Object other ) {
    	if( ! (other instanceof UnifiedClass) ) {
    		return false;
    	}
    	String name = getName();
    	String othername = ((UnifiedClass)other).getName();
    	return name.equals(othername);
    }
    
    @Override
    public String toString() {
        return getName();
    }

    
    // --------------------------------------------------------------------
    // Helper methods
    // --------------------------------------------------------------------
    
    /**
     * Workaround for Class.getName() which for array types, return something
     * like <code>Ljava.lang.Object;</code> (should be <code>Object[]</code>)
     * and for inner types, return something like <code>Outter$Inner</code>
     * (should be Outter.Inner).
     */
    public String getName( Class<?> cl ) {
        
        // Arrays
        if( cl.isArray() ) {
            Class<?> element = cl.getComponentType();     
            element = (Class<?>) resolveGenericType(element);            
            return getName(element)+"[]";
        }
        
        // Inner types
        String name = cl.getName();
        if( name.indexOf('$') != -1 ) {
            name = name.replace('$','.');
        }
        
        return name;
    }

    
    // --------------------------------------------------------------------
    // Implementation specific
    // --------------------------------------------------------------------
    
    /**
     * Return the {@link UClass} corresponding to the specified {@link
     * UnifiedClass} or throw {@link ClassNotFoundException} if this can not be
     * done.
     */
    private UClass toUClass( UnifiedClass uc )
	throws RuntimeClassNotFoundException {
        
        if( uc instanceof UClass ) {
            return (UClass) uc;
        }
        
        /*
         * If the specified UnifiedClass is not an UClass, tries to retrieve a
         * corresponding UClass.
         */
        String name = uc.getName();
        if( jc == null ) {
        	// This instance has been created by constructor UClass(Class<?>)
        	String msg =
        		"Constructor (String,Juliac) should have been invoked for "+
        		"creating this instance instead of constructor (Class<?>)";
        	throw new JuliacRuntimeException(msg);
        }
        UClass other = new UClass(name,jc);
        return other;
    }

    /**
     * Return a string representation of the specified type.
     * 
     * @since 2.1.1
     */
    String toString( Type type ) {
    	
    	/*
    	 * For generic types, retrieve the instantiated value if one has been
    	 * defined.
    	 */
    	type = resolveGenericType(type);
        if( type instanceof Class<?> ) {
            /*
             * For classes, type.toString() returns something like "class
             * MyClass". Calling getName() omits "class".
             */
            String s = getName((Class<?>)type);
            return s;
        }
        else {
        	if( type instanceof ParameterizedType ) {
        		// Generic type
	            ParameterizedType parameterizedType = (ParameterizedType) type;
	            Type rawType = parameterizedType.getRawType();
	            String s = toString(rawType);
	            StringBuilder sb = new StringBuilder(s);
	            Type[] actualTypeArguments =
	            	parameterizedType.getActualTypeArguments();
	            for(int i=0; i<actualTypeArguments.length; i++){
	            	sb.append( i==0 ? '<' : ',' );
	            	s = toString(actualTypeArguments[i]);
	                sb.append(s);
	            }
	            if( actualTypeArguments.length > 0 ){
	            	sb.append('>');
	            }
	            return sb.toString();
        	}
	        else {
	            return type.toString();
	        }
        }
    }
    
    /**
     * If the specified parameter is generic, this method returns its
     * instantiated value if one has been declared. Else this method returns the
     * specified parameter.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
	private final Type resolveGenericType( final Type type ) {
    	Type result = type;
    	if( type instanceof TypeVariable<?> ) {
    		if( genericTypeValues.containsKey((TypeVariable<?>)type) ) {
        		result = genericTypeValues.get(type);    		
    		}
    	}
    	return result;
    }	
}
