/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.ucf;

import java.lang.annotation.Annotation;

/**
 * This interface implements an abstraction for {@link
 * spoon.reflect.declaration.CtType.CtType} and {@link Class}. The idea is to
 * have the possibility of manipulating transparently a type whatever its
 * concrete form (CtType or Class).
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.3
 */
public interface UnifiedClass {

    /**
     * Return the non-primitive (boxed) type which corresponds to the current
     * type. If the current type is non-primitive, return this same type.
     */
    public UnifiedClass box();

    /**
     * Return the instance of the specified type which annotates the current
     * type or <code>null</code> if the current type is not annotated.
     */
    public <A extends Annotation> A getAnnotation( Class<A> annotationType );
    
    /**
     * Return the annotations associated with the current type.
     * 
     * @since 2.2.6
     */
    public Annotation[] getAnnotations();
    
    /**
     * If the class or interface represented by the current type is a
     * member of another class, returns the UnifiedClass object representing the
     * class in which it was declared. This method returns null if this class or
     * interface is not a member of any other class. If this UnifiedClass object
     * represents an array class (currently not implemented), a primitive type,
     * or void, then this method returns null.
     * 
     * @see Class#getDeclaringClass()
     */
    public UnifiedClass getDeclaringClass();
        
    /**
     * Return <code>true</code> if the current type is assignable from the
     * specified UnifiedClass.
     */
    public boolean isAssignableFrom( UnifiedClass uc );
    
    /**
     * Return <code>true</code> if the specified UnifiedClass is assignable from
     * the current type.
     */
    public boolean isAssignableTo( UnifiedClass uc );
    
    /**
     * Return <code>true</code> if the current type is an interface.
     */
    public boolean isInterface();
    
    /**
     * Return <code>true</code> if the current type is primitive.
     */
    public boolean isPrimitive();
    
    /**
     * Return the field, public, protected or private, declared by the current
     * type and corresponding to the specified name.
     * 
     * @param name  the field name
     * @return      the field
     * @throws NoSuchFieldException  if the field does not exist
     */
    public UnifiedField getDeclaredField( String name )
    throws NoSuchFieldException;
    
    /**
     * Return all fields (public, protected and private) fields declared by the
     * current type.
     * 
     * @since 2.2.5
     */
    public UnifiedField[] getDeclaredFields();
    
    /**
     * Return the public fields declared and inherited by the current type.
     * 
     * @since 2.2.5
     */
    public UnifiedField[] getFields();
    
    /**
     * Return the interfaces implemented by the current type.
     * 
     * @since 2.2.5
     */
    public UnifiedClass[] getInterfaces();
    
    /**
     * Return the public methods declared and inherited by the current type.
     */
    public UnifiedMethod[] getMethods();
    
    /**
     * Return the public, declared or inherited, method corresponding to the
     * specified name and parameter types.
     * 
     * @param name   the method name
     * @param types  the parameter types
     * @return       the method
     * @throws NoSuchMethodException  if the method does not exist
     */
    public UnifiedMethod getMethod( String name, Class<?>... types )
    throws NoSuchMethodException;
    
    /**
     * Return the modifiers of the current type.
     */
    public int getModifiers();
    
    /**
     * Return the name of the current type.
     */
    public String getName();
    
    /**
     * Return the null value of the current type. For numerical primitive types,
     * this method returns 0. For booleans, this method returns false. For
     * characters, this method return the space character. For all other types,
     * this method return <code>null</code>.
     */
    public String getNullValue();
    
    /**
     * Return the super class of the current type. If the current type
     * corresponds to an interface, return <code>null</code>.
     * 
     * @since 2.2.5
     */
    public UnifiedClass getSuperclass();
    
    /**
     * Return the string representations of the type variables declared by the
     * current type.
     * 
     * @since 2.1.1
     */
    public String[] getTypeParameters();

    /**
     * Return the string representations of the type variable names declared by
     * the current type. This method differs from {@link #getTypeParameters()}
     * by omitting upper bounds in the return type variables. Upper bounds are
     * elements such as <code>extends FooBar</code> in a generic declaration of
     * the form <code>&lt;T extends FooBar&gt;</code>.
     * 
     * @since 2.1.1
     */
    public String[] getTypeParameterNames();

    /**
     * Return the delimiter for values of the current type. For strings, this
     * method returns ". For characters (char or java.lang.Character), this
     * method returns '. In all other cases, this method returns an empty
     * string.
     * 
     * @since 2.2.4
     */
    public String getValueDelimiter();
}
