/***
 * Juliac
 * Copyright (C) 2010-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Jonathan Labejof
 * Contributor: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.ucf;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Helper methods for dealing with {@link UnifiedClass}es.
 * 
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.5
 */
public class UnifiedClassHelper {

	/**
	 * Return all public declared and inherited methods for the specified class,
	 * provided that, when a method is redefined, only the one with the most
	 * specialized return type is returned.
     */
	public static UnifiedMethod[] getMethodsWithMostSpecializedReturnType( 
		UnifiedClass unifiedClass ) {

		/*
		 * Map of methods to return indexed by their name.
		 */
		Map<String,Collection<UnifiedMethod>> methodsByName =
			new HashMap<String,Collection<UnifiedMethod>>();		

		UnifiedMethod[] methods = unifiedClass.getMethods();
		for(UnifiedMethod method : methods) {

			String methodName = method.getName();
			
			boolean methodAlreadyExists = methodsByName.containsKey(methodName);
			if(!methodAlreadyExists) {
				// Register a new set of method for this name
				Set<UnifiedMethod> set = new HashSet<UnifiedMethod>();
				set.add(method);
				methodsByName.put(methodName,set);
				continue;
			}

			Collection<UnifiedMethod> _methods = methodsByName.get(methodName);
			UnifiedMethod[] _methodsArray = _methods.toArray(new UnifiedMethod[0]);
			for( UnifiedMethod _method : _methodsArray ) {
				
				/*
				 * method and _method share the same name.
				 * Check whether their signature is the same.
				 */
				
				String[] parameters = method.getParameters();
				String[] _parameters = _method.getParameters();
				
				methodAlreadyExists = (parameters.length==_parameters.length);
				if(methodAlreadyExists) {					
					for(int i=0; i<parameters.length; i++) {						
						if(!parameters[i].equals(_parameters[i])){ 
							methodAlreadyExists = false;
							break;
						}
					}
					
					/*
					 * method and _method share the same name and signature.
					 * Check whether the return type of method is more
					 * specialized than the one of _method and in this case,
					 * retain method and discard _method.
					 */
					if(methodAlreadyExists) {
						
						UnifiedClass returnType = method.getReturnType();
						UnifiedClass _returnType = _method.getReturnType();
						if(_returnType.isAssignableFrom(returnType)) {
							_methods.remove(_method);
							_methods.add(method);
						}
						else if(returnType.isAssignableFrom(_returnType)) {
							break;
						}
					}
				}
			}

			/*
			 * At least two different methods with the same name are defined
			 * (declared or inherited) in unifiedClass but their signatures
			 * differ. Register method in the methodsByName map.
			 */
			if(!methodAlreadyExists) {
				_methods.add(method);
			}
		}
		
		// Get all the methods stored in methodsByName
		Collection<UnifiedMethod> unifiedMethods = new HashSet<UnifiedMethod>();
		for(Collection<UnifiedMethod> _methods : methodsByName.values()) {
			for(UnifiedMethod _method : _methods) {
				unifiedMethods.add(_method);
			}
		}
		
		return unifiedMethods.toArray(new UnifiedMethod[0]);		
	}

    /**
     * Return the annotation of type <code>annotationClass</code> associated
     * with <code>uc</code>. If no such annotation exists, recurse into the
     * inheritance hierarchy up to {@link java.lang.Object}. Return
     * <code>null</code> if no such annotation is found.
     */
    public static <A extends Annotation> A getAnnotation(
		UnifiedClass uc, Class<A> annotationClass ) {

    	A annot = uc.getAnnotation(annotationClass);
    	if( annot != null ) {
    		return annot;
    	}
    	
    	// Recurse in the parent class up to java.lang.Object
    	UnifiedClass supercl = uc.getSuperclass();
    	if( ! supercl.getName().equals(Object.class.getName()) ) {
    		annot = getAnnotation(supercl,annotationClass);
    		return annot;
    	}
    	
    	// No such annotation found
    	return null;
    }    

    /**
     * Return all fields declared by the specified class and all fields declared
     * by parent classes, except {@link java.lang.Object}, of the specified
     * class. All fields are included whatever their access modifier is.
     * 
     * @param uc  the class
     * @since 2.2.5.1
     */
    public static UnifiedField[] getAllFields( UnifiedClass uc ) {
    	List<UnifiedField> result = new ArrayList<UnifiedField>();
    	addAllFields(uc,result);
    	return result.toArray(new UnifiedField[result.size()]);
    }
    
    /**
     * Add to the specified list, all fields declared by the specified class and
     * all fields declared by parent classes, except {@link java.lang.Object},
     * of the specified class. All fields are included whatever their access
     * modifier is.
     * 
     * @param uc    the class
     * @param list  the list where fields are to be added
     * @since 2.2.5.1
     */
    public static void addAllFields( UnifiedClass uc, List<UnifiedField> list ) {
    	
    	// Declared fields
    	UnifiedField[] fields = uc.getDeclaredFields();
    	for (UnifiedField field : fields) {
			list.add(field);
		}
    	
    	// Recurse in the parent class up to java.lang.Object
    	UnifiedClass supercl = uc.getSuperclass();
    	if( ! supercl.getName().equals(Object.class.getName()) ) {
        	addAllFields(supercl,list);
    	}
    }    

    /**
     * Return <code>true</code> if the source field overrides the target field.
     * 
     * @param src     the source field
     * @param target  the target field
     * @since 2.2.5.1
     */
    public static boolean override( UnifiedField src, UnifiedField target ) {
    	
    	String srcname = src.getName();
    	String targetname = target.getName();
    	boolean b = srcname.equals(targetname);
    	if(!b) {
    		return false;
    	}
    	
    	UnifiedClass srccl = src.getDeclaringClass();
    	UnifiedClass targetcl = target.getDeclaringClass();
    	
    	if( targetcl.isAssignableFrom(srccl) ) {
    		/*
    		 * If targetcl is assignable from srccl, this means that src (the
    		 * declaring class of) is a subtype of target (idem.) Then, since
    		 * target and src share the same name, src overrides target.
    		 */
    		return true;
    	}
    	
    	return false;
    }
    
    /**
     * Return a copy of the specified array where overriden fields have been
     * removed.
     * 
     * @since 2.2.5.1
     */
    public static UnifiedField[] removeOverriden( UnifiedField[] fields ) {
    	List<UnifiedField> res = new ArrayList<UnifiedField>();
    	res.addAll(Arrays.asList(fields));
    	for (int i = 0; i < fields.length; i++) {
			for (int j = 0; j < fields.length; j++) {
				if(j==i) continue;
				boolean b = override(fields[j],fields[i]);
				if(b) {
					res.remove(fields[i]);
				}
			}
		}
    	return res.toArray(new UnifiedField[res.size()]);
    }
    
    /**
     * Return the annotation associated with the specified unified class among
     * the specified annotation class names. Return <code>null</code> if no such
     * annotation exists.
     * 
     * @param uc               the unified class
     * @param annotClassNames  the annotation class names
     * @return                 the annotation or <code>null</code>
     * @since 2.2.6
     */
    public static Annotation getAnnotation(
		UnifiedClass uc, String... annotClassNames ) {
    	
    	Annotation[] annots = uc.getAnnotations();
    	for (Annotation annot : annots) {
			for (String annotClassName : annotClassNames) {
				String name = annot.annotationType().getName();
				if( name.equals(annotClassName) ) {
					return annot;
				}
			}
		}
    	return null;
    }

    /**
     * Return the annotation associated with the specified unified method among
     * the specified annotation class names. Return <code>null</code> if no such
     * annotation exists.
     * 
     * @param um               the unified method
     * @param annotClassNames  the annotation class names
     * @return                 the annotation or <code>null</code>
     * @since 2.2.6
     */
    public static Annotation getAnnotation(
		UnifiedMethod um, String... annotClassNames ) {
    	
    	Annotation[] annots = um.getAnnotations();
    	for (Annotation annot : annots) {
			for (String annotClassName : annotClassNames) {
				String name = annot.annotationType().getName();
				if( name.equals(annotClassName) ) {
					return annot;
				}
			}
		}
    	return null;
    }
}
