/***
 * Juliac
 * Copyright (C) 2008-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 */

package org.objectweb.fractal.juliac.ucf;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;

/**
 * This class implements an abstraction for {@link
 * spoon.reflect.declaration.CtMethod} and {@link Method}. The idea is to have
 * the possibility of manipulating transparently a method whatever its concrete
 * form (CtMethod or Method).
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @since 1.3
 */
public interface UnifiedMethod {
    
    /**
     * Return the method annotation for the specified annotation type,
     * or <code>null</code> if the method is not annotated.
     */
    public <A extends Annotation> A getAnnotation( Class<A> annotationType );

    /**
     * Return the annotations associated with the current methods.
     * 
     * @since 2.2.6
     */
    public Annotation[] getAnnotations();
    
    /**
     * Return the types of the exceptions thrown by the current method.
     */
    public UnifiedClass[] getExceptionTypes();
    
    /**
     * Return the string representation of the generic exception types thrown by
     * the current method.
     * 
     * @since 2.1
     */
    public String[] getGenericExceptionTypes();
    
    /**
     * Return a string representation of the generic return type for the current
     * method.
     * 
     * @since 2.1
     */
    public String getGenericReturnType();
    
    /**
     * Return the modifiers of the current method.
     */
    public int getModifiers();
    
    /**
     * Return the name of the current method.
     */
    public String getName();
    
    /**
     * Return a string representation (the parameter type followed by the space
     * character, followed by the type name) for the current method.
     */
    public String[] getParameters();
    
    /**
     * Return the types of the parameters for the current method.
     */
    public UnifiedClass[] getParameterTypes();
    
    /**
     * Return the return type of the current method.
     */
    public UnifiedClass getReturnType();    
    
    /**
     * Return the names of the type variables declared by the generic
     * declaration for the current method. Return an empty array if the current
     * method does not declare any type variable.
     * 
     * @since 2.2.6
     */
    public String[] getTypeParameterNames();
    
    /**
     * Return the string representations of the type variables declared by the
     * generic declaration for the current method. Return an empty array if the
     * current method does not declare any type variable.
     * 
     * @since 2.1
     */
    public String[] getTypeParameters();

    /**
     * Return the class declaring the current method.
     * 
     * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
     * @since 2.2.5
     */
	public UnifiedClass getDeclaringClass();
}
