/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.visit;

import static com.google.common.base.Preconditions.checkState;

import java.io.PrintWriter;

/**
 * A {@link FileSourceCodeVisitor} which generates the contains of a source code
 * file.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 * @since 1.1
 */
public class FileSourceCodeWriter extends AbstractCodeWriter
implements FileSourceCodeVisitor {

    public FileSourceCodeWriter( PrintWriter pw ) {
        super(pw);
    }

    /**
     * Visit the source code file header. This header is printed before the
     * package keyword. This header must be a valid Java comment.
     * 
     * @param header  the header comment string
     */
    public void visitFileHeader( String header ) {
    	checkState(state == FILE_HEADER_STATE, "File headers can not be visited at this stage of the visit");

    	writeln(header);
    }
    
    /**
     * Visit the package name.
     * 
     * @param name  the package name
     */
    public void visitPackageName( String name ) {
        if( state < PACKAGE_NAME_STATE ) { state = PACKAGE_NAME_STATE; }
        checkState(state == PACKAGE_NAME_STATE, "Package name can not be visited at this stage of the visit");
        
        writeln();
        writeln("package",name+";");
        writeln();
        
        state = IMPORTS_STATE;
    }
    
    /**
     * Visit an import clause.
     * 
     * @param name  the imported type
     */
    public void visitImport( String name ) {
        if( state < IMPORTS_STATE ) { state = IMPORTS_STATE; }
        checkState(state == IMPORTS_STATE, "Imports can not be visited at this stage of the visit");

        writeln("import",name+";");
    }
    
    /**
     * Visit the public class defined in this file.
     * 
     * @return  a visitor for visiting the source code of the public class
     */
    public ClassSourceCodeVisitor visitPublicClass() {
        if( state < CLASS_STATE ) { state = CLASS_STATE; }
        checkState(state == CLASS_STATE, "Class declaration can not be visited at this stage of the visit");
        
        state = END_STATE;
        return new ClassSourceCodeWriter(writer());        
    }
    
    // ----------------------------------------------------------------------
    // Visitor state constants
    // ----------------------------------------------------------------------
    
    private int state = FILE_HEADER_STATE;
    final static private int FILE_HEADER_STATE = 0;
    final static private int PACKAGE_NAME_STATE = 1;
    final static private int IMPORTS_STATE = 2;
    final static private int CLASS_STATE = 3;
    final static private int END_STATE = 4;
}
